# standard libraries
import copy
import datetime
import io
import json
import os
import pathlib
import typing
import uuid
import zipfile

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift.model import DataItem
from nion.swift.model import Graphics
from nion.swift.model import Utility


class ImportExportIncompatibleDataError(Exception):
    pass


class ImportExportHandler:

    """
        A base class for implementing import/export handlers.

        :param name: the localized name for the handler; will appear in file dialogs
        :param extensions: the list of handled extensions; do not include leading dots
    """

    # Extensions should not include a period.
    def __init__(self, io_handler_id, name, extensions):
        self.io_handler_id = io_handler_id
        self.name = name
        self.extensions = extensions

    def can_read(self):
        return True

    # return data items
    def read_data_items(self, ui, extension, file_path) -> typing.Sequence[DataItem.LibraryItem]:
        data_items = list()
        if os.path.exists(file_path) or file_path.startswith(":"):  # check for colon is for testing
            data_items.extend(self._read_data_items(ui, extension, file_path))
        return data_items

    def _read_data_items(self, ui, extension, file_path) -> typing.Sequence[DataItem.LibraryItem]:
        data_items = list()
        data_elements = self.read_data_elements(ui, extension, file_path)
        for data_element in data_elements:
            if "data" in data_element:
                if not "title" in data_element:
                    root, filename = os.path.split(file_path)
                    title, _ = os.path.splitext(filename)
                    data_element["title"] = title
                data_element["filepath"] = file_path
                data_item = create_data_item_from_data_element(data_element, file_path)
                data_items.append(data_item)
        return data_items

    # return data
    def read_data_elements(self, ui, extension, file_path):
        return list()

    def can_write(self, x_data, extension):
        return False

    def write(self, ui, data_item, path, extension):
        with open(path, 'wb') as f:
            self.write_file(data_item, extension, f)

    def write_file(self, data_item, extension, file):
        data = data_item.data
        if data is not None:
            self.write_data(data, extension, file)

    def write_data(self, data, extension, file):
        pass


class ImportExportManager(metaclass=Utility.Singleton):
    """
        Tracks import/export plugins.
    """
    def __init__(self):
        # we store a of dicts dicts containing extensions,
        # load_func, save_func, keyed by name.
        self.__io_handlers = []

    def register_io_handler(self, io_handler):
        self.__io_handlers.append(io_handler)

    def unregister_io_handler(self, io_handler):
        self.__io_handlers.remove(io_handler)

    def get_readers(self):
        readers = []
        for io_handler in self.__io_handlers:
            if io_handler.can_read():
                readers.append(io_handler)
        return readers

    def get_writers(self):
        writers = []
        for io_handler in self.__io_handlers:
            if io_handler.can_read():
                writers.append(io_handler)
        return writers

    def get_writer_by_id(self, io_handler_id):
        for io_handler in self.__io_handlers:
            if io_handler.io_handler_id == io_handler_id:
                return io_handler
        return None

    def get_writers_for_data_item(self, data_item):
        writers = []
        data_metadata = data_item.data_metadata
        if data_metadata:
            for io_handler in self.__io_handlers:
                for extension in io_handler.extensions:
                    if io_handler.can_write(data_metadata, extension.lower()):
                        writers.append(io_handler)
        return writers

    # read file, return data items
    def read_data_items(self, ui, path):
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                if extension in io_handler.extensions:
                    return io_handler.read_data_items(ui, extension, path)
        return None

    # read file, return data elements
    def read_data_elements(self, ui, path):
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                if extension in io_handler.extensions:
                    return io_handler.read_data_elements(ui, extension, path)
        return None

    def write_data_items_with_writer(self, ui, writer, data_item, path):
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            data_metadata = data_item.data_metadata
            if extension in writer.extensions and data_metadata and writer.can_write(data_metadata, extension):
                writer.write(ui, data_item, path, extension)

    def write_data_items(self, ui, data_item, path):
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                data_metadata = data_item.data_metadata
                if extension in io_handler.extensions and data_metadata and io_handler.can_write(data_metadata, extension):
                    io_handler.write(ui, data_item, path, extension)


# create a new data item with a data element.
# data element is a dict which can be processed into a data item
# when this method returns, the data item has not been added to a document. therefore, the
# data is still loaded into memory, but with a data ref count of zero.
def create_data_item_from_data_element(data_element, data_file_path=None):
    uuid_str = data_element.get("uuid")
    uuid_ = uuid.UUID(uuid_str) if uuid_str else None
    large_format = data_element.get("large_format", False)
    data_item = DataItem.DataItem(item_uuid=uuid_, large_format=large_format)
    update_data_item_from_data_element(data_item, data_element, data_file_path)
    return data_item


# update an existing data item with a data element.
# data element is a dict which can be processed into a data item
# the existing data item may have a new size and dtype after returning.
def update_data_item_from_data_element(data_item, data_element, data_file_path=None):
    data_item.ensure_data_source()
    version = data_element["version"] if "version" in data_element else 1
    if version == 1:
        update_data_item_from_data_element_1(data_item, data_element, data_file_path)
    else:
        raise NotImplementedError("Data element version {:d} not supported.".format(version))


def update_data_item_from_data_element_1(data_item, data_element, data_file_path=None):
    display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
    assert display_specifier.data_item and display_specifier.display
    with data_item.data_item_changes(), display_specifier.data_item.data_source_changes():
        # file path
        # master data
        if data_file_path is not None:
            data_item.source_file_path = data_file_path
        data_and_metadata = convert_data_element_to_data_and_metadata(data_element)
        data = data_and_metadata.data
        dimensional_calibrations = data_and_metadata.dimensional_calibrations
        intensity_calibration = data_and_metadata.intensity_calibration
        is_sequence = data_and_metadata.is_sequence
        collection_dimension_count = data_and_metadata.collection_dimension_count
        datum_dimension_count = data_and_metadata.datum_dimension_count
        data_shape_data_dtype = data_and_metadata.data_shape_and_dtype
        is_same_shape = display_specifier.data_item.data_shape == data_shape_data_dtype[0] and display_specifier.data_item.data_dtype == data_shape_data_dtype[1] and display_specifier.data_item.is_sequence == is_sequence and display_specifier.data_item.collection_dimension_count == collection_dimension_count and display_specifier.data_item.datum_dimension_count == datum_dimension_count
        if is_same_shape:
            with display_specifier.data_item.data_ref() as data_ref:
                sub_area = data_element.get("sub_area")
                if sub_area is not None:
                    top = sub_area[0][0]
                    bottom = sub_area[0][0] + sub_area[1][0]
                    left = sub_area[0][1]
                    right = sub_area[0][1] + sub_area[1][1]
                    data_ref.master_data[top:bottom, left:right] = data[top:bottom, left:right]
                else:
                    data_ref.master_data[:] = data[:]
                data_ref.data_updated()  # trigger change notifications
            if dimensional_calibrations is not None:
                for dimension, dimensional_calibration in enumerate(dimensional_calibrations):
                    display_specifier.data_item.set_dimensional_calibration(dimension, dimensional_calibration)
            if intensity_calibration:
                display_specifier.data_item.set_intensity_calibration(intensity_calibration)
            data_item.metadata = data_and_metadata.metadata
        else:
            display_specifier.data_item.set_xdata(data_and_metadata)
        # title
        if "title" in data_element:
            data_item.title = data_element["title"]
        # description
        # dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
        # time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
        # daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
        # timezone is for conversion and is the Olson timezone string.
        # datetime.datetime.strptime(datetime.datetime.isoformat(datetime.datetime.now()), "%Y-%m-%dT%H:%M:%S.%f" )
        # datetime_modified, datetime_modified_tz, datetime_modified_dst, datetime_modified_tzname is the time at which this image was modified.
        # datetime_original, datetime_original_tz, datetime_original_dst, datetime_original_tzname is the time at which this image was created.
        utc_datetime = data_and_metadata.timestamp
        data_item.created = utc_datetime
        if "time_zone" in data_and_metadata.metadata.get("description", dict()):
            timezone_dict = copy.deepcopy(data_and_metadata.metadata["description"]["time_zone"])
            timezone = timezone_dict.get("timezone")
            if timezone is not None:
                data_item.timezone = timezone
            timezone_offset = timezone_dict.get("tz")
            if timezone_offset is not None:
                data_item.timezone_offset = timezone_offset
        # author
        # sample
        # facility
        # location
        # gps
        # instrument
        # copyright
        # exposure
        # extra_high_tension
        if "arrows" in data_element:
            for arrow_coordinates in data_element["arrows"]:
                start, end = arrow_coordinates
                dimensional_shape = display_specifier.data_item.dimensional_shape
                line_graphic = Graphics.LineGraphic()
                line_graphic.start = (float(start[0]) / dimensional_shape[0], float(start[1]) / dimensional_shape[1])
                line_graphic.end = (float(end[0]) / dimensional_shape[0], float(end[1]) / dimensional_shape[1])
                line_graphic.end_arrow_enabled = True
                display_specifier.display.add_graphic(line_graphic)


def convert_data_element_to_data_and_metadata(data_element) -> DataAndMetadata.DataAndMetadata:
    # NOTE: takes ownership of data_element['data']
    version = data_element["version"] if "version" in data_element else 1
    if version == 1:
        return convert_data_element_to_data_and_metadata_1(data_element)
    else:
        raise NotImplementedError("Data element version {:d} not supported.".format(version))


def convert_data_element_to_data_and_metadata_1(data_element) -> DataAndMetadata.DataAndMetadata:
    """Convert a data element to xdata. No data copying occurs.

    The data element can have the following keys:
        data (required)
        is_sequence, collection_dimension_count, datum_dimension_count (optional description of the data)
        spatial_calibrations (optional list of spatial calibration dicts, scale, offset, units)
        intensity_calibration (optional intensity calibration dict, scale, offset, units)
        metadata (optional)
        properties (get stored into metdata.hardware_source)
        one of either timestamp or datetime_modified
        if datetime_modified (dst, tz) it is converted and used as timestamp
            then timezone gets stored into metadata.description.timezone.
    """
    # data. takes ownership.
    data = data_element["data"]
    dimensional_shape = Image.dimensional_shape_from_data(data)
    is_sequence = data_element.get("is_sequence", False)
    dimension_count = len(Image.dimensional_shape_from_data(data))
    adjusted_dimension_count = dimension_count - (1 if is_sequence else 0)
    collection_dimension_count = data_element.get("collection_dimension_count", 2 if adjusted_dimension_count in (3, 4) else 0)
    datum_dimension_count = data_element.get("datum_dimension_count", adjusted_dimension_count - collection_dimension_count)
    data_descriptor = DataAndMetadata.DataDescriptor(is_sequence, collection_dimension_count, datum_dimension_count)

    # dimensional calibrations
    dimensional_calibrations = None
    if "spatial_calibrations" in data_element:
        dimensional_calibrations_list = data_element.get("spatial_calibrations")
        if len(dimensional_calibrations_list) == len(dimensional_shape):
            dimensional_calibrations = list()
            for dimension_calibration in dimensional_calibrations_list:
                offset = float(dimension_calibration.get("offset", 0.0))
                scale = float(dimension_calibration.get("scale", 1.0))
                units = dimension_calibration.get("units", "")
                units = str(units) if units is not None else str()
                if scale != 0.0:
                    dimensional_calibrations.append(Calibration.Calibration(offset, scale, units))
                else:
                    dimensional_calibrations.append(Calibration.Calibration())

    # intensity calibration
    intensity_calibration = None
    if "intensity_calibration" in data_element:
        intensity_calibration_dict = data_element.get("intensity_calibration")
        offset = float(intensity_calibration_dict.get("offset", 0.0))
        scale = float(intensity_calibration_dict.get("scale", 1.0))
        units = intensity_calibration_dict.get("units", "")
        units = str(units) if units is not None else str()
        if scale != 0.0:
            intensity_calibration = Calibration.Calibration(offset, scale, units)

    # properties (general tags)
    metadata = dict()
    if "metadata" in data_element:
        metadata.update(Utility.clean_dict(data_element.get("metadata")))
    if "properties" in data_element and data_element["properties"]:
        hardware_source_metadata = metadata.setdefault("hardware_source", dict())
        hardware_source_metadata.update(Utility.clean_dict(data_element.get("properties")))

    # dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
    # time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
    # daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
    # timezone is for conversion and is the Olson timezone string.
    # datetime.datetime.strptime(datetime.datetime.isoformat(datetime.datetime.now()), "%Y-%m-%dT%H:%M:%S.%f" )
    # datetime_modified, datetime_modified_tz, datetime_modified_dst, datetime_modified_tzname is the time at which this image was modified.
    # datetime_original, datetime_original_tz, datetime_original_dst, datetime_original_tzname is the time at which this image was created.
    timestamp = data_element.get("timestamp", datetime.datetime.utcnow())
    datetime_item = data_element.get("datetime_modified", Utility.get_datetime_item_from_utc_datetime(timestamp))

    local_datetime = Utility.get_datetime_from_datetime_item(datetime_item)
    dst_value = datetime_item.get("dst", "+00")
    tz_value = datetime_item.get("tz", "+0000")
    timezone = datetime_item.get("timezone")
    time_zone = { "dst": dst_value, "tz": tz_value}
    if timezone is not None:
        time_zone["timezone"] = timezone
    # note: dst is informational only; tz already include dst
    tz_adjust = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
    utc_datetime = local_datetime - datetime.timedelta(minutes=tz_adjust)  # tz_adjust already contains dst_adjust
    timestamp = utc_datetime
    metadata.setdefault("description", dict())["time_zone"] = time_zone

    return DataAndMetadata.new_data_and_metadata(data, intensity_calibration, dimensional_calibrations, metadata=metadata, timestamp=timestamp, data_descriptor=data_descriptor)


def create_data_element_from_data_item(data_item, include_data=True):
    data_element = dict()
    data_element["version"] = 1
    data_element["reader_version"] = 1
    if data_item.has_data:
        if include_data:
            data_element["data"] = data_item.data
        dimensional_calibrations = data_item.dimensional_calibrations
        if dimensional_calibrations is not None:
            calibrations_element = list()
            for calibration in dimensional_calibrations:
                calibration_element = { "offset": calibration.offset, "scale": calibration.scale, "units": calibration.units }
                calibrations_element.append(calibration_element)
            data_element["spatial_calibrations"] = calibrations_element
        intensity_calibration = data_item.intensity_calibration
        if intensity_calibration is not None:
            intensity_calibration_element = { "offset": intensity_calibration.offset, "scale": intensity_calibration.scale, "units": intensity_calibration.units }
            data_element["intensity_calibration"] = intensity_calibration_element
        if data_item.is_sequence:
            data_element["is_sequence"] = data_item.is_sequence
        data_element["collection_dimension_count"] = data_item.collection_dimension_count
        data_element["datum_dimension_count"] = data_item.datum_dimension_count
        data_element["metadata"] = copy.deepcopy(data_item.metadata)
        data_element["properties"] = copy.deepcopy(data_item.metadata.get("hardware_source", dict()))
        data_element["title"] = data_item.title
        data_element["source_file_path"] = data_item.source_file_path
        tz_value = data_item.timezone_offset
        timezone = data_item.timezone
        dst_minutes = None
        time_zone_dict = data_item.metadata.get("description", dict()).get("time_zone")
        if time_zone_dict:
            # note: dst is informational only; tz already include dst
            if tz_value is None:
                tz_value = time_zone_dict["tz"]
            dst_minutes = int(time_zone_dict["dst"])
            if timezone is None:
                timezone = time_zone_dict.get("timezone")
        if tz_value is not None:
            tz_minutes = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
        data_element["datetime_modified"] = Utility.get_datetime_item_from_utc_datetime(data_item.created, tz_minutes, dst_minutes, timezone)
        data_element["datetime_original"] = Utility.get_datetime_item_from_utc_datetime(data_item.created, tz_minutes, dst_minutes, timezone)
        data_element["uuid"] = str(data_item.uuid)
        # operation
        # graphics
    return data_element


def create_data_element_from_extended_data(xdata: DataAndMetadata.DataAndMetadata) -> dict:
    data_element = dict()
    data_element["version"] = 1
    data_element["reader_version"] = 1
    data_element["data"] = xdata.data
    data_element["large_format"] = len(xdata.dimensional_shape) > 2
    dimensional_calibrations = xdata.dimensional_calibrations
    calibrations_element = list()
    for calibration in dimensional_calibrations:
        calibration_element = { "offset": calibration.offset, "scale": calibration.scale, "units": calibration.units }
        calibrations_element.append(calibration_element)
    data_element["spatial_calibrations"] = calibrations_element
    intensity_calibration = xdata.intensity_calibration
    intensity_calibration_element = { "offset": intensity_calibration.offset, "scale": intensity_calibration.scale, "units": intensity_calibration.units }
    data_element["intensity_calibration"] = intensity_calibration_element
    if xdata.is_sequence:
        data_element["is_sequence"] = xdata.is_sequence
    data_element["collection_dimension_count"] = xdata.collection_dimension_count
    data_element["datum_dimension_count"] = xdata.datum_dimension_count
    data_element["metadata"] = copy.deepcopy(xdata.metadata)
    # properties is redundant; but here for backwards compatibility
    data_element["properties"] = copy.deepcopy(xdata.metadata.get("hardware_source", dict()))
    tz_minutes = None
    dst_minutes = None
    timezone = None
    time_zone_dict = xdata.metadata.get("description", dict()).get("time_zone")
    if time_zone_dict:
        # note: dst is informational only; tz already include dst
        tz_value = time_zone_dict["tz"]
        tz_minutes = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
        dst_minutes = int(time_zone_dict["dst"])
        timezone = time_zone_dict.get("timezone")
    data_element["datetime_modified"] = Utility.get_datetime_item_from_utc_datetime(xdata.timestamp, tz_minutes, dst_minutes, timezone)
    data_element["datetime_original"] = Utility.get_datetime_item_from_utc_datetime(xdata.timestamp, tz_minutes, dst_minutes, timezone)
    return data_element


class StandardImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id, name, extensions):
        super(StandardImportExportHandler, self).__init__(io_handler_id, name, extensions)

    def read_data_elements(self, ui, extension, path):
        data = None
        try:
            data = Image.read_image_from_file(ui, path)
        except Exception as e:
            pass
        if data is not None:
            data_element = dict()
            data_element["version"] = 1
            data_element["data"] = data
            if os.path.exists(path) or path.startswith(":"):  # check for colon is for testing
                try:
                    file_datetime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                except:
                    file_datetime = None
                if file_datetime is not None:
                    data_element["datetime_modified"] = Utility.get_datetime_item_from_datetime(file_datetime)
            return [data_element]
        return list()

    def can_write(self, data_and_metadata, extension):
        return len(data_and_metadata.dimensional_shape) == 2

    def write(self, ui, data_item, path, extension):
        display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
        data = display_specifier.display.get_calculated_display_values(True).display_rgba  # export the display rather than the data for these types
        if data is not None:
            ui.save_rgba_data_to_file(data, path, extension)


class CSVImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id, name, extensions):
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, ui, extension, path):
        data = numpy.loadtxt(path, delimiter=',')
        if data is not None:
            data_element = dict()
            data_element["data"] = data
            return [data_element]
        return list()

    def can_write(self, x_data, extension):
        return True

    def write(self, ui, data_item, path, extension):
        data = data_item.data
        if data is not None:
            numpy.savetxt(path, data, delimiter=', ')


class CSV1ImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id, name, extensions):
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, ui, extension, path):
        return None

    def can_write(self, x_data, extension):
        return x_data and x_data.is_data_1d

    def write(self, ui, data_item, path, extension):
        data = data_item.data
        calibration = data_item.xdata.dimensional_calibrations[0]
        intensity_calibration = data_item.xdata.intensity_calibration
        if data is not None:
            calibrated_data = numpy.empty(data.shape + (2,), data.dtype)
            length = calibrated_data.shape[0]
            calibrated_data[:, 0] = numpy.linspace(calibration.offset, calibration.offset + (length - 1) * calibration.scale, length)
            calibrated_data[:, 1] = intensity_calibration.offset + data * intensity_calibration.scale
            header = str(calibration.units) + ", " + str(intensity_calibration.units)
            numpy.savetxt(path, calibrated_data, delimiter=', ', header=header)


class NDataImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id, name, extensions):
        super(NDataImportExportHandler, self).__init__(io_handler_id, name, extensions)

    def read_data_elements(self, ui, extension, path):
        zip_file = zipfile.ZipFile(path, 'r')
        namelist = zip_file.namelist()
        if "metadata.json" in namelist and "data.npy" in namelist:
            metadata = json.loads(zip_file.read("metadata.json").decode("utf-8"))
            data_bytes = zip_file.read("data.npy")
            data = numpy.load(io.BytesIO(data_bytes))
        if data is not None:
            data_element = metadata
            data_element["data"] = data
            return [data_element]
        return list()

    def can_write(self, data_and_metadata, extension):
        return True

    def write(self, ui, data_item, path, extension):
        data_element = create_data_element_from_data_item(data_item, include_data=False)
        data = data_item.data
        if data is not None:
            root, ext = os.path.splitext(path)
            metadata_path = root + "_metadata.json"
            data_path = root + "_data.npy"
            try:
                with open(metadata_path, "w") as fp:
                    json.dump(data_element, fp)
                numpy.save(data_path, data)
                zip_file = zipfile.ZipFile(path, 'w')
                zip_file.write(metadata_path, "metadata.json")
                zip_file.write(data_path, "data.npy")
            finally:
                os.remove(metadata_path)
                os.remove(data_path)


class NumPyImportExportHandler(ImportExportHandler):
    """A file import/export handler to read/write the npy file type.

    The npy file type is the one included in the NumPy package.

    This i/o handler will optionally read metadata with the same name but the
    '.json' suffix located in the same directory.

    This i/o handler will write metadata to a file with the same name in the same
    directory but with the '.json' suffix.
    """

    def __init__(self, io_handler_id, name, extensions):
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, ui, extension: str, path_str: str) -> typing.List[dict]:
        path = pathlib.Path(path_str)
        data = numpy.load(str(path))
        metadata_path = path.with_suffix(".json")
        if metadata_path.exists():
            metadata = json.load(metadata_path)
        else:
            metadata = dict()
        if data is not None:
            data_element = metadata
            data_element["data"] = data
            return [data_element]
        return list()

    def can_write(self, data_and_metadata, extension: str) -> bool:
        return True

    def write(self, ui, data_item: DataItem.DataItem, path_str: str, extension: str) -> None:
        data_path = pathlib.Path(path_str)
        metadata_path = data_path.with_suffix(".json")
        data_element = create_data_element_from_data_item(data_item, include_data=False)
        data = data_item.data
        if data is not None:
            try:
                with open(str(metadata_path), "w") as fp:
                    json.dump(data_element, fp)
                numpy.save(str(data_path), data)
            except Exception:
                os.remove(str(metadata_path))
                os.remove(str(data_path))
                raise


# Register the intrinsic I/O handlers.
ImportExportManager().register_io_handler(StandardImportExportHandler("jpeg-io-handler", "JPEG", ["jpg", "jpeg"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("png-io-handler", "PNG", ["png"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("gif-io-handler", "GIF", ["gif"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("bmp-io-handler", "BMP", ["bmp"]))
ImportExportManager().register_io_handler(CSVImportExportHandler("csv-io-handler", "CSV Raw", ["csv"]))
ImportExportManager().register_io_handler(CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"]))
ImportExportManager().register_io_handler(NDataImportExportHandler("ndata1-io-handler", "NData 1", ["ndata1"]))
ImportExportManager().register_io_handler(NumPyImportExportHandler("numpy-io-handler", "Raw NumPy", ["npy"]))



# standard libraries
import binascii
import contextlib
import datetime
import io
import json
import logging
import os
import shutil
import unittest
import uuid

# third party libraries
import numpy

# local libraries
from nion.swift.model import NDataHandler
from nion.swift.model import Cache


class TestNDataHandlerClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_ndata_handler_basic_functionality(self):
        now = datetime.datetime.now()
        current_working_directory = os.getcwd()
        data_dir = os.path.join(current_working_directory, "__Test")
        Cache.db_make_directory_if_needed(data_dir)
        try:
            h = NDataHandler.NDataHandler(os.path.join(data_dir, "abc.ndata"))
            with contextlib.closing(h):
                p = {u"abc": 1, u"def": u"bcd", u"uuid": str(uuid.uuid4())}
                # write properties
                h.write_properties(p, now)
                self.assertEqual(h.read_properties(), p)
                self.assertIsNone(h.read_data())
                # write data
                h.write_data(numpy.zeros((4,4), dtype=numpy.float64), now)
                self.assertEqual(h.read_properties(), p)
                d = h.read_data()
                self.assertEqual(d.shape, (4, 4))
                self.assertEqual(d.dtype, numpy.float64)
                # rewrite data
                h.write_data(numpy.zeros((12,12), dtype=numpy.float32), now)
                self.assertEqual(h.read_properties(), p)
                d = h.read_data()
                self.assertEqual(d.shape, (12, 12))
                self.assertEqual(d.dtype, numpy.float32)
                # rewrite properties
                h.write_properties(p, now)
                self.assertEqual(h.read_properties(), p)
                d = h.read_data()
                self.assertEqual(d.shape, (12, 12))
                self.assertEqual(d.dtype, numpy.float32)
        finally:
            #logging.debug("rmtree %s", data_dir)
            shutil.rmtree(data_dir)

    def test_ndata_handler_rewrites_reversed_zip_file(self):
        now = datetime.datetime.now()
        current_working_directory = os.getcwd()
        data_dir = os.path.join(current_working_directory, "__Test")
        Cache.db_make_directory_if_needed(data_dir)
        try:
            p = {u"abc": 1, u"def": u"bcd", u"uuid": str(uuid.uuid4())}
            d = numpy.zeros((12,12), dtype=numpy.float32)
            # write zip file where metadata is first
            with open(os.path.join(data_dir, "file.ndata"), "w+b") as fp:
                dir_data_list = list()
                dt = now
                properties = p
                data = d
                if properties is not None:
                    json_io = io.StringIO()
                    json.dump(properties, json_io)
                    json_str = json_io.getvalue()
                    def write_json(fp):
                        json_bytes = bytes(json_str, 'ISO-8859-1')
                        fp.write(json_bytes)
                        return binascii.crc32(json_bytes) & 0xFFFFFFFF
                    offset_json = fp.tell()
                    json_len, json_crc32 = NDataHandler.write_local_file(fp, b"metadata.json", write_json, dt)
                    dir_data_list.append((offset_json, b"metadata.json", json_len, json_crc32))
                if data is not None:
                    offset_data = fp.tell()
                    def write_data(fp):
                        numpy_start_pos = fp.tell()
                        numpy.save(fp, data)
                        numpy_end_pos = fp.tell()
                        fp.seek(numpy_start_pos)
                        header_data = fp.read((numpy_end_pos - numpy_start_pos) - data.nbytes)  # read the header
                        data_crc32 = binascii.crc32(data.data, binascii.crc32(header_data)) & 0xFFFFFFFF
                        fp.seek(numpy_end_pos)
                        return data_crc32
                    data_len, crc32 = NDataHandler.write_local_file(fp, b"data.npy", write_data, dt)
                    dir_data_list.append((offset_data, b"data.npy", data_len, crc32))
                dir_offset = fp.tell()
                for offset, name_bytes, data_len, crc32 in dir_data_list:
                    NDataHandler.write_directory_data(fp, offset, name_bytes, data_len, crc32, dt)
                dir_size = fp.tell() - dir_offset
                NDataHandler.write_end_of_directory(fp, dir_size, dir_offset, len(dir_data_list))
                fp.truncate()
            # make sure read works
            h = NDataHandler.NDataHandler(os.path.join(data_dir, "file.ndata"))
            with contextlib.closing(h):
                self.assertEqual(h.read_properties(), p)
                dd = h.read_data()
                self.assertEqual(dd.shape, d.shape)
                self.assertEqual(dd.dtype, d.dtype)
                # now rewrite
                h.write_properties(p, now)
        finally:
            #logging.debug("rmtree %s", data_dir)
            shutil.rmtree(data_dir)

    def test_ndata_handles_discontiguous_data(self):
        logging.getLogger().setLevel(logging.DEBUG)
        now = datetime.datetime.now()
        current_working_directory = os.getcwd()
        data_dir = os.path.join(current_working_directory, "__Test")
        Cache.db_make_directory_if_needed(data_dir)
        try:
            h = NDataHandler.NDataHandler(os.path.join(data_dir, "abc.ndata"))
            with contextlib.closing(h):
                data = numpy.random.randint(0, 10, size=(10, 10))[:,3]  # discontiguous data
                self.assertFalse(data.flags['C_CONTIGUOUS'])
                p = {u"uuid": str(uuid.uuid4())}
                # write properties
                h.write_properties(p, now)
                # write data
                h.write_data(data, now)
                d = h.read_data()
                self.assertEqual(d.shape, data.shape)
                self.assertEqual(d.dtype, data.dtype)
        finally:
            #logging.debug("rmtree %s", data_dir)
            shutil.rmtree(data_dir)

    def test_ndata_handles_corrupt_data(self):
        logging.getLogger().setLevel(logging.DEBUG)
        now = datetime.datetime.now()
        current_working_directory = os.getcwd()
        data_dir = os.path.join(current_working_directory, "__Test")
        Cache.db_make_directory_if_needed(data_dir)
        try:
            zero_path = os.path.join(data_dir, "zeros.ndata")
            with open(zero_path, 'wb') as f:
                f.write(bytearray(1024))
            with self.assertRaises(IOError):
                with open(zero_path, "rb") as fp:
                    NDataHandler.parse_zip(fp)
        finally:
            #logging.debug("rmtree %s", data_dir)
            shutil.rmtree(data_dir)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



# standard libraries
import gettext
import numpy
import time
import typing

# third party libraries
# None

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Undo
from nion.ui import Dialog
from nion.utils import Binding
from nion.utils import Converter
from nion.utils import Geometry
from nion.utils import Model
from nion.swift import DataItemThumbnailWidget
from nion.swift.model import DataItem

_ = gettext.gettext


class Recorder:

    def __init__(self, document_controller, data_item):
        self.__document_controller = document_controller
        self.__document_model = document_controller.document_model
        self.__data_item = data_item

        self.on_live_state_changed = None
        self.on_recording_state_changed = None
        self.on_data_item_removed = None

        self.__recording_state = "stopped"
        self.__recording_start = 0.0
        self.__recording_index = 0
        self.__recording_data_item = None
        self.__recording_transaction = None
        self.__recording_error = False
        self.__recording_interval = None
        self.__recording_count = None

        def data_item_content_changed():
            is_live = data_item.is_live
            if not is_live:
                self.__stop_recording()
            if callable(self.on_live_state_changed):
                self.on_live_state_changed(is_live)

        # if the item source or recorded data item is removed, stop recording
        # called when an item is removed from the document
        def item_removed(key, value, index):
            if value == self.__recording_data_item:
                self.__stop_recording()
            if value == self.__data_item:
                self.__stop_recording()
                if callable(self.on_data_item_removed):
                    self.on_data_item_removed()

        def data_changed():
            current_xdata = self.__data_item.xdata
            if current_xdata and not current_xdata.is_sequence:
                hardware_source_metadata = current_xdata.metadata.get("hardware_source", dict())
                valid_rows = hardware_source_metadata.get("valid_rows", 0) if "valid_rows" in hardware_source_metadata else current_xdata.data_shape[0]
                if len(current_xdata.data_shape) == 1 or valid_rows >= current_xdata.data_shape[0]:
                    self.__last_complete_xdata = current_xdata
            else:
                self.__recording_error = True

        self.__library_item_changed_event_listener = data_item.library_item_changed_event.listen(data_item_content_changed)
        self.__item_removed_event_listener = self.__document_model.item_removed_event.listen(item_removed)
        self.__data_item_data_changed_event_listener = self.__data_item.data_changed_event.listen(data_changed)

        self.__last_complete_xdata = None

        data_item_content_changed()
        data_changed()

    def close(self):
        self.__library_item_changed_event_listener.close()
        self.__library_item_changed_event_listener = None
        self.__item_removed_event_listener.close()
        self.__item_removed_event_listener = None
        self.__data_item_data_changed_event_listener.close()
        self.__data_item_data_changed_event_listener = None

    @property
    def recording_state(self) -> str:
        return self.__recording_state

    class RecorderInsertLibraryItemCommandCommand(Undo.UndoableCommand):

        def __init__(self, document_controller, recorder: "Recorder", library_item_fn: typing.Callable[[], DataItem.LibraryItem]):
            super().__init__(_("Record Library Item"))
            self.__document_controller = document_controller
            self.__recorder = recorder
            self.__old_workspace_layout = self.__document_controller.workspace_controller.deconstruct()
            self.__new_workspace_layout = None
            self.__library_item_uuid = None
            self.__library_item_fn = library_item_fn
            self.__library_item_index = None
            self.initialize()

        def close(self):
            self.__document_controller = None
            self.__library_item_uuid = None
            self.__library_item_fn = None
            self.__library_item_index = None
            self.__old_workspace_layout = None
            self.__new_workspace_layout = None
            super().close()

        def perform(self):
            self.__library_item_uuid = self.__library_item_fn().uuid

        @property
        def library_item(self):
            return self.__document_controller.document_model.get_data_item_by_uuid(self.__library_item_uuid)

        def _get_modified_state(self):
            return self.__document_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__document_controller.document_model.modified_state = modified_state

        def _redo(self):
            self.__document_controller.document_model.undelete_all(self.__undelete_log)
            self.__library_item_uuid = self.__document_controller.document_model.data_items[self.__library_item_index].uuid
            self.__document_controller.workspace_controller.reconstruct(self.__new_workspace_layout)

        def _undo(self):
            library_item = self.__document_controller.document_model.get_data_item_by_uuid(self.__library_item_uuid)
            self.__recorder.stop_recording()
            self.__new_workspace_layout = self.__document_controller.workspace_controller.deconstruct()
            self.__library_item_index = self.__document_controller.document_model.data_items.index(library_item)
            self.__undelete_log = self.__document_controller.document_model.remove_data_item(library_item, safe=True)
            self.__document_controller.workspace_controller.reconstruct(self.__old_workspace_layout)

    def continue_recording(self, current_time: float) -> None:
        if self.__recording_state == "recording":
            if (current_time - self.__recording_start) // self.__recording_interval > self.__recording_index - 1:
                self.__recording_last = current_time
                # first create an empty data item to hold the recorded data if it doesn't already exist
                if not self.__recording_data_item:
                    data_item = DataItem.DataItem(large_format=True)
                    data_item.ensure_data_source()
                    data_item.title = _("Recording of ") + self.__data_item.title

                    def process():
                        self.__document_model.append_data_item(data_item)
                        self.__document_controller.display_data_item(DataItem.DisplaySpecifier.from_data_item(data_item))
                        return data_item

                    command = Recorder.RecorderInsertLibraryItemCommandCommand(self.__document_controller, self, process)
                    command.perform()
                    self.__document_controller.push_undo_command(command)

                    self.__recording_data_item = data_item
                # next grab the current data and stop if it is a sequence (can't record sequences)
                current_xdata = self.__last_complete_xdata
                if self.__recording_error:
                    self.__stop_recording()
                    return
                if not current_xdata:
                    # no first image yet
                    return
                # now record the new data. it may or may not be a new frame at this point.
                last_xdata = self.__recording_data_item.xdata
                self.__recording_index += 1
                if current_xdata and last_xdata and current_xdata.data_shape == self.__recording_data_item.data_shape[1:]:
                    # continue, append the new data to existing data and update existing data item
                    intensity_calibration = last_xdata.intensity_calibration
                    dimensional_calibrations = last_xdata.dimensional_calibrations
                    data_descriptor = last_xdata.data_descriptor
                    sequence_xdata = DataAndMetadata.new_data_and_metadata(
                        numpy.vstack([last_xdata.data, [current_xdata.data]]),
                        intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations,
                        data_descriptor=data_descriptor)
                    self.__recording_data_item.set_xdata(sequence_xdata)
                elif current_xdata and not last_xdata:
                    # first acquisition, create the sequence
                    intensity_calibration = current_xdata.intensity_calibration
                    dimensional_calibrations = [Calibration.Calibration(scale=self.__recording_interval,
                                                                        units="s")] + list(
                        current_xdata.dimensional_calibrations)
                    data_descriptor = DataAndMetadata.DataDescriptor(True,
                                                                     current_xdata.data_descriptor.collection_dimension_count,
                                                                     current_xdata.data_descriptor.datum_dimension_count)
                    sequence_xdata = DataAndMetadata.new_data_and_metadata(current_xdata.data[numpy.newaxis, ...],
                                                                           intensity_calibration=intensity_calibration,
                                                                           dimensional_calibrations=dimensional_calibrations,
                                                                           data_descriptor=data_descriptor)
                    self.__recording_data_item.set_xdata(sequence_xdata)
                    self.__recording_transaction = self.__document_model.item_transaction(self.__recording_data_item)
                else:
                    # something is amiss. stop.
                    self.__stop_recording()
                    return
            # finally -- check if we've reached the maximum count
            if self.__recording_index >= self.__recording_count:
                self.__stop_recording()

    def start_recording(self, recording_start: float, recording_interval: float, recording_count: int) -> None:
        self.__recording_state = "recording"
        self.__recording_start = recording_start
        self.__recording_index = 0
        self.__recording_error = False
        self.__recording_interval = recording_interval
        self.__recording_count = recording_count
        if callable(self.on_recording_state_changed):
            self.on_recording_state_changed(self.__recording_state)

    def stop_recording(self):
        self.__stop_recording()

    def __stop_recording(self):
        if self.__recording_state == "recording":
            self.__recording_state = "stopped"
            self.__recording_start = 0.0
            self.__recording_index = 0
            self.__recording_error = False
            if self.__recording_data_item and self.__recording_transaction:
                self.__recording_transaction.close()
                self.__recording_transaction = None
                self.__recording_data_item = None
            if callable(self.on_recording_state_changed):
                self.on_recording_state_changed(self.__recording_state)


class RecorderDialog(Dialog.ActionDialog):

    def __init__(self, document_controller, data_item):
        ui = document_controller.ui
        super().__init__(ui, _("Recorder"), app=document_controller.app, parent_window=document_controller, persistent_id="Recorder" + str(data_item.uuid))

        self.__recorder = Recorder(document_controller, data_item)

        self.ui = ui
        self.document_controller = document_controller

        self.__data_item = data_item

        self._create_menus()

        self.__record_button = ui.create_push_button_widget(_("Record"))

        def thumbnail_widget_drag(mime_data, thumbnail, hot_spot_x, hot_spot_y):
            # use this convoluted base object for drag so that it doesn't disappear after the drag.
            self.content.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)

        data_item_thumbnail_source = DataItemThumbnailWidget.DataItemThumbnailSource(ui, data_item=data_item)
        data_item_chooser_widget = DataItemThumbnailWidget.ThumbnailWidget(ui, data_item_thumbnail_source, Geometry.IntSize(48, 48))
        data_item_chooser_widget.on_drag = thumbnail_widget_drag

        self.__recording_interval_property = Model.PropertyModel(1000)
        self.__recording_count_property = Model.PropertyModel(20)

        recording_period_widget = ui.create_line_edit_widget(properties={"width": 60})
        recording_period_widget.bind_text(Binding.PropertyBinding(self.__recording_interval_property, "value", converter=Converter.IntegerToStringConverter()))

        recording_count_widget = ui.create_line_edit_widget(properties={"width": 60})
        recording_count_widget.bind_text(Binding.PropertyBinding(self.__recording_count_property, "value", converter=Converter.IntegerToStringConverter()))

        row0 = ui.create_row_widget()
        row0.add_stretch()
        row0.add_spacing(8)
        row0.add(self.__record_button)
        row0.add_spacing(8)

        row1 = ui.create_row_widget()
        row1.add(ui.create_label_widget(_("Interval")))
        row1.add_spacing(8)
        row1.add(recording_period_widget)
        row1.add_spacing(4)
        row1.add(ui.create_label_widget(_("msec")))
        row1.add_spacing(8)
        row1.add_stretch()

        row2 = ui.create_row_widget()
        row2.add(ui.create_label_widget(_("Frames")))
        row2.add_spacing(8)
        row2.add(recording_count_widget)
        row2.add_spacing(8)
        row2.add_stretch()

        column1 = ui.create_column_widget()
        column1.add(row1)
        column1.add_spacing(4)
        column1.add(row2)
        column1.add_spacing(4)
        column1.add(row0)

        button_row = ui.create_row_widget()
        button_row.add_spacing(8)
        button_row.add(data_item_chooser_widget)
        button_row.add_spacing(8)
        button_row.add_stretch()
        button_row.add_spacing(8)
        button_row.add(column1)
        button_row.add_spacing(8)

        def record_pressed():
            if self.__recorder.recording_state == "recording":
                self.__recorder.stop_recording()
            else:
                self.__recorder.start_recording(time.time(), self.__recording_interval_property.value / 1000, self.__recording_count_property.value)

        self.__record_button.on_clicked = record_pressed

        column = self.content
        column.add_spacing(6)
        column.add(button_row)
        column.add_spacing(6)

        def live_state_changed(is_live: bool) -> None:
            self.__record_button.enabled = is_live

        def recording_state_changed(recording_state: str) -> None:
            if recording_state == "recording":
                self.__record_button.text = _("Stop")
            else:
                self.__record_button.text = _("Record")

        def data_item_removed() -> None:
            self.request_close()

        self.__recorder.on_live_state_changed = live_state_changed
        self.__recorder.on_recording_state_changed = recording_state_changed
        self.__recorder.on_data_item_removed = data_item_removed

        live_state_changed(data_item.is_live)
        recording_state_changed("stopped")

    def close(self):
        self.__recorder.close()
        self.__recorder = None
        super().close()

    def periodic(self):
        super().periodic()
        self.__recorder.continue_recording(time.time())



"""
    Contains classes related to thumbnail generation.
"""

# standard libraries
import threading
import time

# third-party libraries
import numpy

# local libraries
from nion.swift import DisplayPanel
from nion.swift.model import Utility
from nion.swift.model.Display import Display
from nion.ui import DrawingContext
from nion.utils import Event
from nion.utils import ReferenceCounting


class ThumbnailProcessor:
    """Processes thumbnails for a display in a thread."""

    def __init__(self, display):
        self.__display = display
        self.__cache = display._display_cache
        self.__cache_property_name = "thumbnail_data"
        # the next two fields represent a memory cache -- a cache of the cache values.
        # if self.__cached_value_dirty is None then this first level cache has not yet
        # been initialized. these fields are used for optimization.
        self.__cached_value = None
        self.__cached_value_dirty = None
        self.__cached_value_time = 0
        self.__is_recomputing = False
        self.__is_recomputing_lock = threading.RLock()
        self.width = 72
        self.height = 72
        self.on_thumbnail_updated = None
        self.__recompute_thread = None
        self.__recompute_lock = threading.RLock()
        self.__recompute_thread_cancel = threading.Event()

    def close(self):
        self.on_thumbnail_updated = None
        with self.__is_recomputing_lock:
            if self.__recompute_thread:
                self.__recompute_thread_cancel.set()
                self.__recompute_thread.join()

    # used for testing
    @property
    def _is_cached_value_dirty(self):
        return self.__cached_value_dirty

    # thread safe
    def mark_data_dirty(self):
        """ Called from item to indicate its data or metadata has changed."""
        self.__cache.set_cached_value_dirty(self.__display, self.__cache_property_name)
        self.__initialize_cache()
        self.__cached_value_dirty = True

    def __initialize_cache(self):
        """Initialize the cache values (cache values are used for optimization)."""
        if self.__cached_value_dirty is None:
            self.__cached_value_dirty = self.__cache.is_cached_value_dirty(self.__display, self.__cache_property_name)
            self.__cached_value = self.__cache.get_cached_value(self.__display, self.__cache_property_name)

    def recompute_if_necessary(self, ui):
        """Recompute the data on a thread, if necessary.

        If the data has recently been computed, this call will be rescheduled for the future.

        If the data is currently being computed, it do nothing."""
        self.__initialize_cache()
        if self.__cached_value_dirty:
            with self.__is_recomputing_lock:
                is_recomputing = self.__is_recomputing
                self.__is_recomputing = True
            if is_recomputing:
                pass
            else:
                # the only way to get here is if we're not currently computing
                # this has the side effect of limiting the number of threads that
                # are sleeping.
                def recompute():
                    try:
                        if self.__recompute_thread_cancel.wait(0.01):  # helps tests run faster
                            return
                        minimum_time = 0.5
                        current_time = time.time()
                        if current_time < self.__cached_value_time + minimum_time:
                            if self.__recompute_thread_cancel.wait(self.__cached_value_time + minimum_time - current_time):
                                return
                        self.recompute_data(ui)
                    finally:
                        self.__is_recomputing = False
                        self.__recompute_thread = None
                with self.__is_recomputing_lock:
                    self.__recompute_thread = threading.Thread(target=recompute)
                    self.__recompute_thread.start()

    def recompute_data(self, ui):
        """Compute the data associated with this processor.

        This method is thread safe and may take a long time to return. It should not be called from
         the UI thread. Upon return, the results will be calculated with the latest data available
         and the cache will not be marked dirty.
        """
        self.__initialize_cache()
        with self.__recompute_lock:
            if self.__cached_value_dirty:
                try:
                    calculated_data = self.get_calculated_data(ui)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    traceback.print_stack()
                    raise
                self.__cache.set_cached_value(self.__display, self.__cache_property_name, calculated_data)
                self.__cached_value = calculated_data
                self.__cached_value_dirty = False
                self.__cached_value_time = time.time()
            else:
                calculated_data = None
            if calculated_data is None:
                calculated_data = self.get_default_data()
                if calculated_data is not None:
                    # if the default is not None, treat is as valid cached data
                    self.__cache.set_cached_value(self.__display, self.__cache_property_name, calculated_data)
                    self.__cached_value = calculated_data
                    self.__cached_value_dirty = False
                    self.__cached_value_time = time.time()
                else:
                    # otherwise remove everything from the cache
                    self.__cache.remove_cached_value(self.__display, self.__cache_property_name)
                    self.__cached_value = None
                    self.__cached_value_dirty = None
                    self.__cached_value_time = 0
            self.__recompute_lock.release()
            if callable(self.on_thumbnail_updated):
                self.on_thumbnail_updated()
            self.__recompute_lock.acquire()

    def get_data(self, ui):
        """Return the computed data for this processor.

        This method is thread safe but may take a long time to return since it may have to compute
         the results. It should not be called from the UI thread.
        """
        self.recompute_data(ui)
        return self.get_cached_data()

    def get_cached_data(self):
        """Return the cached data for this processor.

        This method is thread safe and always returns quickly, using the cached data.
        """
        self.__initialize_cache()
        return self.__cached_value

    def get_calculated_data(self, ui):
        drawing_context = DisplayPanel.preview(ui, self.__display, 512, 512)
        thumbnail_drawing_context = DrawingContext.DrawingContext()
        thumbnail_drawing_context.scale(self.width / 512, self.height / 512)
        thumbnail_drawing_context.add(drawing_context)
        return ui.create_rgba_image(thumbnail_drawing_context, self.width, self.height)

    def get_default_data(self):
        return numpy.zeros((self.height, self.width), dtype=numpy.uint32)


class ThumbnailSource(ReferenceCounting.ReferenceCounted):
    """Produce a thumbnail for a display."""

    def __init__(self, ui, display: Display):
        super().__init__()
        self._ui = ui
        self._display = display

        self.thumbnail_updated_event = Event.Event()
        self.__thumbnail_processor = ThumbnailProcessor(display)
        self._on_will_delete = None

        def thumbnail_changed():
            thumbnail_processor = self.__thumbnail_processor
            if thumbnail_processor:
                thumbnail_processor.mark_data_dirty()
                thumbnail_processor.recompute_if_necessary(ui)

        no_cached_data = self.__thumbnail_processor.get_cached_data() is None

        self.__next_calculated_display_values_listener = display.add_calculated_display_values_listener(thumbnail_changed, send=no_cached_data)
        self.__display_changed_event_listener = display.display_changed_event.listen(thumbnail_changed)

        def thumbnail_updated():
            self.thumbnail_updated_event.fire()

        self.__thumbnail_processor.on_thumbnail_updated = thumbnail_updated

        self.__thumbnail_processor.recompute_if_necessary(ui)

        def display_will_close():
            if self.__thumbnail_processor:
                self.__thumbnail_processor.close()
                self.__thumbnail_processor = None

        self.__display_will_close_listener = display.about_to_be_removed_event.listen(display_will_close)

    def close(self):
        self.remove_ref()

    def about_to_delete(self):
        self.__display_will_close_listener.close()
        self.__display_will_close_listener = None
        if self.__thumbnail_processor:
            self.__thumbnail_processor.close()
            self.__thumbnail_processor = None
        if self.__display_changed_event_listener:
            self.__display_changed_event_listener.close()
            self.__display_changed_event_listener = None
        if self.__next_calculated_display_values_listener:
            self.__next_calculated_display_values_listener.close()
            self.__next_calculated_display_values_listener = None
        self._on_will_delete(self)
        self._on_will_delete = None

    @property
    def thumbnail_data(self):
        return self.__thumbnail_processor.get_cached_data() if self.__thumbnail_processor else None

    def recompute_data(self):
        self.__thumbnail_processor.recompute_data(self._ui)

    # used for testing
    @property
    def _is_thumbnail_dirty(self):
        return self.__thumbnail_processor._is_cached_value_dirty


class ThumbnailManager(metaclass=Utility.Singleton):
    """Manages thumbnail sources for displays."""

    def __init__(self):
        self.__thumbnail_sources = dict()
        self.__lock = threading.RLock()

    def thumbnail_source_for_display(self, ui, display: Display) -> ThumbnailSource:
        """Returned ThumbnailSource must be closed."""
        with self.__lock:
            thumbnail_source = self.__thumbnail_sources.get(display)
            if not thumbnail_source:
                thumbnail_source = ThumbnailSource(ui, display)
                self.__thumbnail_sources[display] = thumbnail_source

                def will_delete(thumbnail_source):
                    del self.__thumbnail_sources[thumbnail_source._display]

                thumbnail_source._on_will_delete = will_delete
            else:
                assert thumbnail_source._ui == ui
            return thumbnail_source.add_ref()

    def thumbnail_data_for_display(self, display: Display) -> numpy.ndarray:
        thumbnail_source = self.__thumbnail_sources.get(display) if display else None
        if thumbnail_source:
            return thumbnail_source.thumbnail_data
        return None



# standard libraries
import asyncio
import collections
import contextlib
import datetime
import functools
import logging
import sys
import threading
import time
import traceback

# third party libraries
import numpy

# local libraries
# None


# datetimes are _local_ datetimes and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
# time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
# daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
# timezone is for conversion and is the Olson timezone string.
# datetime_item is a dictionary with entries for the local_datetime, tz (timezone offset), and
# dst (daylight savings time offset). it may optionally include tz_name (timezone name), if available.
def get_datetime_item_from_datetime(datetime_local, tz_minutes=None, dst_minutes=None, timezone=None):
    # dst is information, tz already includes dst
    datetime_item = dict()
    datetime_item["local_datetime"] = datetime_local.isoformat()
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - datetime.datetime.utcnow()).total_seconds())) // 60
    if dst_minutes is None:
        dst_minutes = 60 if time.localtime().tm_isdst else 0
    datetime_item["tz"] = '{0:+03d}{1:02d}'.format(tz_minutes // 60, tz_minutes % 60)
    datetime_item["dst"] = "+{0:02d}".format(dst_minutes)
    if timezone is not None:
        datetime_item["timezone"] = timezone
    return datetime_item


def get_current_datetime_item():
    return get_datetime_item_from_datetime(datetime.datetime.now())


def get_datetime_item_from_utc_datetime(datetime_utc, tz_minutes=None, dst_minutes=None, timezone=None):
    # dst is information, tz already includes dst
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - datetime.datetime.utcnow()).total_seconds())) // 60
    return get_datetime_item_from_datetime(datetime_utc + datetime.timedelta(minutes=tz_minutes), tz_minutes, dst_minutes, timezone)

local_utcoffset_override = None  # for testing
try:
    import pytz.reference
    def local_utcoffset_minutes(datetime_local:datetime.datetime=None) -> int:
        if local_utcoffset_override is not None:
            return local_utcoffset_override[0]
        datetime_local = datetime_local if datetime_local else datetime.datetime.now()
        return int(pytz.reference.LocalTimezone().utcoffset(datetime_local).total_seconds() // 60)
except ImportError:
    def local_utcoffset_minutes(datetime_local:datetime.datetime=None) -> int:
        if local_utcoffset_override is not None:
            return local_utcoffset_override[0]
        return int(round((datetime.datetime.now() - datetime.datetime.utcnow()).total_seconds())) // 60

class TimezoneMinutesToStringConverter:
    def convert(self, value):
        return "{0:+03d}{1:02d}".format(value // 60, value % 60) if value is not None else None
    def convert_back(self, value):
        return (int(value[1:3]) * 60 + int(value[3:5])) * (-1 if value[0] == '-' else 1) if value is not None else None

local_timezone_override = None  # for testing
try:
    import tzlocal
    def get_local_timezone():
        return tzlocal.get_localzone().zone if local_timezone_override is None else local_timezone_override[0]
except ImportError:
    def get_local_timezone():
        return None if local_timezone_override is None else local_timezone_override[0]

# return python datetime object from a datetime_item. may return None if the datetime element is
# not properly formatted.
def get_datetime_from_datetime_item(datetime_item):
    local_datetime = datetime_item.get("local_datetime", str())
    if len(local_datetime) == 26:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S.%f")
    elif len(local_datetime) == 19:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S")
    return None


class Singleton(type):
    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


def clean_dict(d0, clean_item_fn=None):
    """
        Return a json-clean dict. Will log info message for failures.
    """
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    d = dict()
    for key in d0:
        cleaned_item = clean_item_fn(d0[key])
        if cleaned_item is not None:
            d[key] = cleaned_item
    return d


def clean_list(l0, clean_item_fn=None):
    """
        Return a json-clean list. Will log info message for failures.
    """
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l = list()
    for index, item in enumerate(l0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return l


def clean_tuple(t0, clean_item_fn=None):
    """
        Return a json-clean tuple. Will log info message for failures.
    """
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l = list()
    for index, item in enumerate(t0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return tuple(l)


def clean_item(i):
    """
        Return a json-clean item or None. Will log info message for failure.
    """
    itype = type(i)
    if itype == dict:
        return clean_dict(i)
    elif itype == list:
        return clean_list(i)
    elif itype == tuple:
        return clean_tuple(i)
    elif itype == numpy.float32:
        return float(i)
    elif itype == numpy.float64:
        return float(i)
    elif itype == numpy.int16:
        return int(i)
    elif itype == numpy.uint16:
        return int(i)
    elif itype == numpy.int32:
        return int(i)
    elif itype == numpy.uint32:
        return int(i)
    elif itype == numpy.int64:
        return int(i)
    elif itype == numpy.uint64:
        return int(i)
    elif itype == float:
        return i
    elif itype == str:
        return i
    elif itype == int:
        return i
    elif itype == bool:
        return i
    elif itype == type(None):
        return i
    logging.info("[1] Unable to handle type %s", itype)
    import traceback
    traceback.print_stack()
    return None


def clean_item_no_list(i):
    """
        Return a json-clean item or None. Will log info message for failure.
    """
    itype = type(i)
    if itype == dict:
        return clean_dict(i, clean_item_no_list)
    elif itype == list:
        return clean_tuple(i, clean_item_no_list)
    elif itype == tuple:
        return clean_tuple(i, clean_item_no_list)
    elif itype == numpy.float32:
        return float(i)
    elif itype == numpy.float64:
        return float(i)
    elif itype == numpy.int16:
        return int(i)
    elif itype == numpy.uint16:
        return int(i)
    elif itype == numpy.int32:
        return int(i)
    elif itype == numpy.uint32:
        return int(i)
    elif itype == float:
        return i
    elif itype == str:
        return i
    elif itype == int:
        return i
    elif itype == bool:
        return i
    elif itype == type(None):
        return i
    logging.info("[2] Unable to handle type %s", itype)
    return None


def parse_version(version, count=3, max_count=None):
    max_count = max_count if max_count is not None else count
    version_components = [int(version_component) for version_component in version.split(".")]
    assert len(version_components) <= max_count
    while len(version_components) < count:
        version_components.append(0)
    return version_components


def compare_versions(version1: str, version2: str) -> int:
    if version1.startswith("~"):
        version1 = version1[1:]
        version_components1 = parse_version(version1, 1, 3)
        assert len(version_components1) > 1
    elif version1 == "1":  # same as "~1.0"
        version1 = "1.0"
        version_components1 = parse_version(version1, 2, 3)
    else:
        version_components1 = parse_version(version1)
    version_components2 = parse_version(version2)
    # print(version_components1, version_components2)
    for version_component1, version_component2 in zip(version_components1, version_components2):
        if version_component1 > version_component2:
            return 1
        elif version_component1 < version_component2:
            return -1
    return 0


def fps_tick(fps_id):
    v = globals().setdefault("__fps_" + fps_id, [0, 0.0, None, 0.0, None, []])
    v[0] += 1
    next_time = time.perf_counter()
    while len(v[5]) > 100:
        v[5].pop(0)
    if v[2] is not None:
        v[5].append(next_time - v[2])
    v[1] += next_time - v[2] if v[2] is not None else 0.0
    if v[1] > 1.0:
        v[3] = v[0] / v[1]
        v[0] = 0
        v[1] = 0.0
    v[2] = next_time
    return fps_get(fps_id)

def fps_get(fps_id):
    v = globals().setdefault("__fps_" + fps_id, [0, 0.0, None, 0.0, None, []])
    # s = numpy.std(v[5]) if len(v[5]) > 0 else 0.0
    return str(int(v[3]*100)/100.0) # + " " + str(int(s*1000)) + "ms"

def trace_calls(trace, frame, event, arg):
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    t = time.time()
    last_elapsed = t - trace.last_time_ref[0]
    if not trace.discard or trace.discard not in func_filename:
        if last_elapsed > trace.min_elapsed:
            s = "         " if last_elapsed > 0.001 else ""
            trace.all.append(":%s%12.5f %12.5f Call to %s on line %s of %s from line %s of %s" % (
                s, last_elapsed, t - trace.start_time, func_name, func_line_no, func_filename, caller_line_no, caller_filename))
        trace.last_time_ref[0] = t


def begin_trace(min_elapsed, discard):
    Trace = collections.namedtuple("Trace", ["start_time", "last_time_ref", "min_elapsed", "discard", "all"])
    trace = Trace(start_time=time.time(), last_time_ref=[0], min_elapsed=min_elapsed, discard=discard, all=list())
    sys.settrace(functools.partial(trace_calls, trace))
    return trace


def end_trace(trace):
    sys.settrace(None)
    logging.debug("\n".join(trace.all))
    logging.debug("TOTAL: %s", len(trace.all))


@contextlib.contextmanager
def trace(min_elapsed=0.0, discard=None):
    t = begin_trace(min_elapsed, discard)
    yield
    end_trace(t)


def sample_stack_all(count=10, interval=0.1):
    """Sample the stack in a thread and print it at regular intervals."""

    def print_stack_all(l, ll):
        l1 = list()
        l1.append("*** STACKTRACE - START ***")
        code = []
        for threadId, stack in sys._current_frames().items():
            sub_code = []
            sub_code.append("# ThreadID: %s" % threadId)
            for filename, lineno, name, line in traceback.extract_stack(stack):
                sub_code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    sub_code.append("  %s" % (line.strip()))
            if not "in select" in sub_code[-2] and \
               not "in wait" in sub_code[-2] and \
               not "in print_stack_all" in sub_code[-2] and \
               not "in sample_stack_all" in sub_code[-2] and \
               not "in checkcache" in sub_code[-2] and \
               not "do_sleep" in sub_code[-2] and \
               not "sleep" in sub_code[-1] and \
               not any(["in do_sample" in s for s in sub_code]):
                code.extend(sub_code)
        for line in code:
            l1.append(line)
        l1.append("*** STACKTRACE - END ***")
        with l:
            ll.extend(l1)

    def do_sample():
        l = threading.RLock()
        ll = list()
        for i in range(count):
            print_stack_all(l, ll)
            time.sleep(interval)
        with l:
            print("\n".join(ll))

    threading.Thread(target=do_sample).start()


class TestEventLoop:
    def __init__(self, event_loop: asyncio.AbstractEventLoop = None):
        logging.disable(logging.CRITICAL)  # suppress new_event_loop debug message
        self.__event_loop = event_loop if event_loop else asyncio.new_event_loop()
        logging.disable(logging.NOTSET)
        self.__event_loop.has_no_pulse = True

    def close(self):
        # give cancelled tasks a chance to finish
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        self.__event_loop.run_until_complete(asyncio.gather(*asyncio.Task.all_tasks(loop=self.__event_loop), loop=self.__event_loop))
        # now close
        # due to a bug in Python libraries, the default executor needs to be shutdown explicitly before the event loop
        # see http://bugs.python.org/issue28464
        if self.__event_loop._default_executor:
            self.__event_loop._default_executor.shutdown()
        self.__event_loop.close()
        self.__event_loop = None

    @property
    def event_loop(self):
        return self.__event_loop



# -*- coding: utf-8 -*-

import setuptools
import os

setuptools.setup(
    name="nionswift",
    version="0.13.0",
    packages=["nion.swift", "nion.swift.model", "nion.swift.test", "nionui_app.nionswift", "nionswift_plugin.none", "nionlib", "nion.typeshed"],
    package_data={"nion.swift": ["resources/*"]},
    install_requires=['scipy', 'numpy', 'h5py', 'pytz', 'tzlocal', 'pillow', 'nionutils', 'niondata', 'nionui', 'nionswift-io'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha"
    ],
    include_package_data=True,
    test_suite="nion.swift.test",
    entry_points={
        'console_scripts': [
            'nionswift=nion.swift.command:main',
            ],
        },
)



# standard libraries
import copy
import logging
import math
import threading
import typing

# third party libraries
import numpy
import scipy.ndimage

# local libraries
from nion.data import Calibration
from nion.swift import Undo
from nion.swift.model import Graphics
from nion.swift.model import Utility
from nion.ui import CanvasItem
from nion.utils import Geometry


class ImageCanvasItemMapping:

    def __init__(self, data_shape, canvas_origin, canvas_size):
        assert data_shape is None or len(data_shape) == 2
        self.data_shape = data_shape
        # double check dimensions are not zero
        if self.data_shape:
            for d in self.data_shape:
                if not d > 0:
                    self.data_shape = None
        # calculate transformed image rect
        self.canvas_rect = None
        if self.data_shape:
            rect = (canvas_origin, canvas_size)
            self.canvas_rect = Geometry.fit_to_size(rect, self.data_shape)

    def map_point_image_norm_to_widget(self, p):
        p = Geometry.FloatPoint.make(p)
        if self.data_shape:
            return Geometry.FloatPoint(y=p.y * self.canvas_rect.height + self.canvas_rect.top, x=p.x * self.canvas_rect.width + self.canvas_rect.left)
        return None

    def map_size_image_norm_to_widget(self, s):
        ms = self.map_point_image_norm_to_widget(s)
        ms0 = self.map_point_image_norm_to_widget((0, 0))
        return ms - ms0

    def map_size_image_to_image_norm(self, s):
        ms = self.map_point_image_to_image_norm(s)
        ms0 = self.map_point_image_to_image_norm((0, 0))
        return ms - ms0

    def map_size_image_to_widget(self, s):
        ms = self.map_point_image_to_widget(s)
        ms0 = self.map_point_image_to_widget((0, 0))
        return ms - ms0

    def map_size_widget_to_image_norm(self, s):
        ms = self.map_point_widget_to_image_norm(s)
        ms0 = self.map_point_widget_to_image_norm((0, 0))
        return ms - ms0

    def map_point_widget_to_image_norm(self, p):
        if self.data_shape:
            p = Geometry.FloatPoint.make(p)
            p_image = self.map_point_widget_to_image(p)
            return Geometry.FloatPoint(y=p_image.y / self.data_shape[0], x=p_image.x / self.data_shape[1])
        return None

    def map_point_widget_to_image(self, p):
        if self.canvas_rect and self.data_shape:
            p = Geometry.FloatPoint.make(p)
            if self.canvas_rect.height != 0.0:
                image_y = self.data_shape[0] * (p.y - self.canvas_rect.top) / self.canvas_rect.height
            else:
                image_y = 0.0
            if self.canvas_rect.width != 0.0:
                image_x = self.data_shape[1] * (p.x - self.canvas_rect.left) / self.canvas_rect.width
            else:
                image_x = 0.0
            return Geometry.FloatPoint(y=image_y, x=image_x)  # c-indexing
        return None

    def map_point_image_norm_to_image(self, p):
        if self.data_shape:
            p = Geometry.FloatPoint.make(p)
            return Geometry.FloatPoint(y=p.y * self.data_shape[0], x=p.x * self.data_shape[1])
        return None

    def map_point_image_to_image_norm(self, p):
        if self.data_shape:
            p = Geometry.FloatPoint.make(p)
            return Geometry.FloatPoint(y=p.y / self.data_shape[0], x=p.x / self.data_shape[1])
        return None

    def map_point_image_to_widget(self, p):
        p = Geometry.FloatPoint.make(p)
        if self.data_shape:
            return Geometry.FloatPoint(y=p.y * self.canvas_rect.height / self.data_shape[0] + self.canvas_rect.top, x=p.x * self.canvas_rect.width / self.data_shape[1] + self.canvas_rect.left)
        return None


class GraphicsCanvasItem(CanvasItem.AbstractCanvasItem):
    """A canvas item to paint the graphic items on the image.

    Callers should call update_graphics when the graphics changes.
    """

    def __init__(self, get_font_metrics_fn):
        super(GraphicsCanvasItem, self).__init__()
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__displayed_shape = None
        self.__graphics = None
        self.__graphics_for_compare = list()
        self.__graphic_selection = None

    def update_graphics(self, displayed_shape, graphics, graphic_selection):
        if displayed_shape is None or len(displayed_shape) != 2:
            displayed_shape = None
            graphics = None
            graphic_selection = None
        assert displayed_shape is None or len(displayed_shape) == 2
        needs_update = False
        if ((self.__displayed_shape is None) != (displayed_shape is None)) or (self.__displayed_shape != displayed_shape):
            self.__displayed_shape = displayed_shape
            needs_update = True
        graphics_for_compare = [graphic.mime_data_dict() for graphic in (graphics or list())]
        if graphics_for_compare != self.__graphics_for_compare:
            self.__graphics = graphics
            self.__graphics_for_compare = graphics_for_compare
            needs_update = True
        if self.__graphic_selection != graphic_selection:
            self.__graphic_selection = graphic_selection
            needs_update = True
        if needs_update:
            self.update()

    def _repaint(self, drawing_context):
        if self.__graphics:
            widget_mapping = ImageCanvasItemMapping(self.__displayed_shape, (0, 0), self.canvas_size)
            with drawing_context.saver():
                for graphic_index, graphic in enumerate(self.__graphics):
                    if isinstance(graphic, (Graphics.PointTypeGraphic, Graphics.LineTypeGraphic, Graphics.RectangleTypeGraphic, Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                        try:
                            graphic.draw(drawing_context, self.__get_font_metrics_fn, widget_mapping, self.__graphic_selection.contains(graphic_index))
                        except Exception as e:
                            import traceback
                            logging.debug("Graphic Repaint Error: %s", e)
                            traceback.print_exc()
                            traceback.print_stack()


class InfoOverlayCanvasItem(CanvasItem.AbstractCanvasItem):
    """A canvas item to paint the scale marker as an overlay.

    Callers should set the image_canvas_origin and image_canvas_size properties.

    Callers should also call set_data_info when the data changes.
    """

    def __init__(self):
        super(InfoOverlayCanvasItem, self).__init__()
        self.__image_canvas_size = None  # this will be updated by the container
        self.__image_canvas_origin = None  # this will be updated by the container
        self.__data_shape = None
        self.__dimensional_calibration = None
        self.__info_text = None

    @property
    def _dimension_calibration_for_test(self):
        return self.__dimensional_calibration

    @property
    def image_canvas_size(self):
        return self.__image_canvas_size

    @image_canvas_size.setter
    def image_canvas_size(self, value):
        if self.__image_canvas_size is None or value != self.__image_canvas_size:
            self.__image_canvas_size = value
            self.update()

    @property
    def image_canvas_origin(self):
        return self.__image_canvas_origin

    @image_canvas_origin.setter
    def image_canvas_origin(self, value):
        if self.__image_canvas_origin is None or value != self.__image_canvas_origin:
            self.__image_canvas_origin = value
            self.update()

    def set_data_info(self, data_shape, dimensional_calibration: Calibration.Calibration, metadata: dict) -> None:
        needs_update = False
        if self.__data_shape is None or data_shape != self.__data_shape:
            self.__data_shape = data_shape
            needs_update = True
        if self.__dimensional_calibration is None or dimensional_calibration != self.__dimensional_calibration:
            self.__dimensional_calibration = dimensional_calibration
            needs_update = True
        info_items = list()
        hardware_source_metadata = metadata.get("hardware_source", dict())
        voltage = hardware_source_metadata.get("autostem", dict()).get("high_tension_v", 0)
        if voltage:
            units = "V"
            if voltage % 1000 == 0:
                voltage = voltage // 1000
                units = "kV"
            info_items.append("{0} {1}".format(voltage, units))
        hardware_source_name = hardware_source_metadata.get("hardware_source_name")
        if hardware_source_name:
            info_items.append(str(hardware_source_name))
        info_text = " ".join(info_items)
        if self.__info_text is None or self.__info_text != info_text:
            self.__info_text = info_text
            needs_update = True
        if needs_update:
            self.update()

    def _repaint(self, drawing_context):
        canvas_size = self.canvas_size
        canvas_height = canvas_size[0]
        image_canvas_size = self.image_canvas_size
        image_canvas_origin = self.image_canvas_origin
        dimensional_calibration = self.__dimensional_calibration
        if dimensional_calibration is not None and image_canvas_origin is not None and image_canvas_size is not None:  # display scale marker?
            origin = (canvas_height - 30, 20)
            scale_marker_width = 120
            scale_marker_height = 6
            data_shape = self.__data_shape
            widget_mapping = ImageCanvasItemMapping(data_shape, image_canvas_origin, image_canvas_size)
            if data_shape[0] > 1.0 and data_shape[1] > 0.0:
                screen_pixel_per_image_pixel = widget_mapping.map_size_image_norm_to_widget((1, 1))[0] / data_shape[0]
                if screen_pixel_per_image_pixel > 0:
                    scale_marker_image_width = scale_marker_width / screen_pixel_per_image_pixel
                    calibrated_scale_marker_width = Geometry.make_pretty2(scale_marker_image_width * dimensional_calibration.scale, True)
                    # update the scale marker width
                    scale_marker_image_width = calibrated_scale_marker_width / dimensional_calibration.scale
                    scale_marker_width = scale_marker_image_width * screen_pixel_per_image_pixel
                    with drawing_context.saver():
                        drawing_context.begin_path()
                        drawing_context.move_to(origin[1], origin[0])
                        drawing_context.line_to(origin[1] + scale_marker_width, origin[0])
                        drawing_context.line_to(origin[1] + scale_marker_width, origin[0] - scale_marker_height)
                        drawing_context.line_to(origin[1], origin[0] - scale_marker_height)
                        drawing_context.close_path()
                        drawing_context.fill_style = "#448"
                        drawing_context.fill()
                        drawing_context.stroke_style = "#000"
                        drawing_context.stroke()
                        drawing_context.font = "normal 14px serif"
                        drawing_context.text_baseline = "bottom"
                        drawing_context.fill_style = "#FFF"
                        drawing_context.fill_text(dimensional_calibration.convert_to_calibrated_size_str(scale_marker_image_width), origin[1], origin[0] - scale_marker_height - 4)
                        drawing_context.fill_text(self.__info_text, origin[1], origin[0] - scale_marker_height - 4 - 20)


class ImageCanvasItemDelegate:
    # interface must be implemented by the delegate

    def begin_mouse_tracking(self) -> None: ...

    def end_mouse_tracking(self, undo_command) -> None: ...

    def delete_key_pressed(self) -> None: ...

    def enter_key_pressed(self) -> None: ...

    def cursor_changed(self, pos): ...

    def update_display_properties(self, display_properties: typing.Mapping) -> None: ...

    def create_insert_graphics_command(self, graphics: typing.Sequence[Graphics.Graphic]) -> Undo.UndoableCommand: ...

    def create_change_display_command(self, *, command_id: str=None, is_mergeable: bool=False) -> Undo.UndoableCommand: ...

    def create_change_graphics_command(self) -> Undo.UndoableCommand: ...

    def push_undo_command(self, command: Undo.UndoableCommand) -> None: ...

    def add_index_to_selection(self, index: int) -> None: ...

    def remove_index_from_selection(self, index: int) -> None: ...

    def set_selection(self, index: int) -> None: ...

    def clear_selection(self) -> None: ...

    def nudge_selected_graphics(self, mapping, delta) -> None: ...

    def drag_graphics(self, graphics) -> None: ...

    def update_graphics(self, widget_mapping, graphic_drag_items, graphic_drag_part, graphic_part_data, graphic_drag_start_pos, pos, modifiers) -> None: ...

    def image_clicked(self, image_position, modifiers) -> bool: ...

    def image_mouse_pressed(self, image_position, modifiers) -> bool: ...

    def image_mouse_released(self, image_position, modifiers) -> bool: ...

    def image_mouse_position_changed(self, image_position, modifiers) -> bool: ...

    def show_display_context_menu(self, gx, gy) -> bool: ...

    def create_rectangle(self, pos): ...

    def create_ellipse(self, pos): ...

    def create_line(self, pos): ...

    def create_point(self, pos): ...

    def create_line_profile(self, pos): ...

    def create_spot(self, pos): ...

    def create_wedge(self, angle): ...

    def create_ring(self, radius): ...

    @property
    def tool_mode(self) -> str: return str()

    @tool_mode.setter
    def tool_mode(self, value: str) -> None: ...


def calculate_origin_and_size(canvas_size, data_shape, image_canvas_mode, image_zoom, image_position) -> typing.Tuple[typing.Any, typing.Any]:
    """Calculate origin and size for canvas size, data shape, and image display parameters."""
    if data_shape is None:
        return None, None
    if image_canvas_mode == "fill":
        data_shape = data_shape
        scale_h = float(data_shape[1]) / canvas_size[1]
        scale_v = float(data_shape[0]) / canvas_size[0]
        if scale_v < scale_h:
            image_canvas_size = (canvas_size[0], canvas_size[0] * data_shape[1] / data_shape[0])
        else:
            image_canvas_size = (canvas_size[1] * data_shape[0] / data_shape[1], canvas_size[1])
        image_canvas_origin = (canvas_size[0] * 0.5 - image_canvas_size[0] * 0.5, canvas_size[1] * 0.5 - image_canvas_size[1] * 0.5)
    elif image_canvas_mode == "fit":
        image_canvas_size = canvas_size
        image_canvas_origin = (0, 0)
    elif image_canvas_mode == "1:1":
        image_canvas_size = data_shape
        image_canvas_origin = (canvas_size[0] * 0.5 - image_canvas_size[0] * 0.5, canvas_size[1] * 0.5 - image_canvas_size[1] * 0.5)
    elif image_canvas_mode == "2:1":
        image_canvas_size = (data_shape[0] * 0.5, data_shape[1] * 0.5)
        image_canvas_origin = (canvas_size[0] * 0.5 - image_canvas_size[0] * 0.5, canvas_size[1] * 0.5 - image_canvas_size[1] * 0.5)
    else:
        image_canvas_size = (canvas_size[0] * image_zoom, canvas_size[1] * image_zoom)
        canvas_rect = Geometry.fit_to_size(((0, 0), image_canvas_size), data_shape)
        image_canvas_origin_y = (canvas_size[0] * 0.5) - image_position[0] * canvas_rect[1][0] - canvas_rect[0][0]
        image_canvas_origin_x = (canvas_size[1] * 0.5) - image_position[1] * canvas_rect[1][1] - canvas_rect[0][1]
        image_canvas_origin = (image_canvas_origin_y, image_canvas_origin_x)
    return image_canvas_origin, image_canvas_size


class ImageCanvasItem(CanvasItem.LayerCanvasItem):
    """A canvas item to paint an image.

    Callers are expected to pass in a delegate.

    They are expected to call the following functions to update the display:
        update_image_display_state
        update_regions

    The delegate is expected to handle the following events:
        add_index_to_selection(index)
        remove_index_from_selection(index)
        set_selection(index)
        clear_selection()
        nudge_selected_graphics(mapping, delta)
        update_graphics(widget_mapping, graphic_drag_items, graphic_drag_part, graphic_part_data, graphic_drag_start_pos, pos, modifiers)
        tool_mode (property)
        show_display_context_menu(gx, gy)
        begin_mouse_tracking(self)
        end_mouse_tracking()
        mouse_clicked(image_position, modifiers)
        delete_key_pressed()
        enter_key_pressed()
        cursor_changed(pos)
    """

    def __init__(self, get_font_metrics_fn, delegate: ImageCanvasItemDelegate, event_loop, draw_background: bool=True):
        super().__init__()

        self.__get_font_metrics_fn = get_font_metrics_fn
        self.delegate = delegate
        self.__event_loop = event_loop

        self.wants_mouse_events = True

        self.__update_layout_handle = None
        self.__update_layout_handle_lock = threading.RLock()

        self.__closing_lock = threading.RLock()
        self.__closed = False

        self.__image_zoom = 1.0
        self.__image_position = (0.5, 0.5)
        self.__image_canvas_mode = "fit"

        self.__last_display_values = None
        self.__graphics_changed = False

        # create the child canvas items
        # the background
        # next the zoomable items
        self.__bitmap_canvas_item = CanvasItem.BitmapCanvasItem(background_color="#888" if draw_background else "transparent")
        self.__graphics_canvas_item = GraphicsCanvasItem(get_font_metrics_fn)
        self.__timestamp_canvas_item = CanvasItem.TimestampCanvasItem()
        # put the zoomable items into a composition
        self.__composite_canvas_item = CanvasItem.CanvasItemComposition()
        self.__composite_canvas_item.add_canvas_item(self.__bitmap_canvas_item)
        self.__composite_canvas_item.add_canvas_item(self.__graphics_canvas_item)
        self.__composite_canvas_item.add_canvas_item(self.__timestamp_canvas_item)
        # and put the composition into a scroll area
        self.scroll_area_canvas_item = CanvasItem.ScrollAreaCanvasItem(self.__composite_canvas_item)
        self.scroll_area_canvas_item._constrain_position = False  # temporary until scroll bars are implemented

        def layout_updated(canvas_origin, canvas_size, *, immediate=False):
            self.__update_overlay_canvas_item(canvas_size, immediate=immediate)

        self.scroll_area_canvas_item.on_layout_updated = layout_updated
        # info overlay (scale marker, etc.)
        self.__info_overlay_canvas_item = InfoOverlayCanvasItem()
        # canvas items get added back to front
        if draw_background:
            self.add_canvas_item(CanvasItem.BackgroundCanvasItem())
        self.add_canvas_item(self.scroll_area_canvas_item)
        self.add_canvas_item(self.__info_overlay_canvas_item)

        self.__display_values = None
        self.__data_shape = None
        self.__graphics = list()
        self.__graphic_selection = None

        # used for tracking undo
        self.__undo_command = None

        # used for dragging graphic items
        self.__graphic_drag_items = []
        self.__graphic_drag_item = None
        self.__graphic_part_data = {}
        self.__graphic_drag_indexes = []
        self.__last_mouse = None
        self.__is_dragging = False
        self.__mouse_in = False

        # frame rate and latency
        self.__display_frame_rate_id = None
        self.__display_frame_rate_last_index = 0
        self.__display_latency = False

    def close(self):
        with self.__closing_lock:
            with self.__update_layout_handle_lock:
                update_layout_handle = self.__update_layout_handle
                if update_layout_handle:
                    update_layout_handle.cancel()
                    self.__update_layout_handle = None
            self.__closed = True
        super().close()

    @property
    def default_aspect_ratio(self):
        return 1.0

    @property
    def _info_overlay_canvas_item_for_test(self):
        return self.__info_overlay_canvas_item

    def display_inserted(self, display, index):
        pass

    def display_removed(self, display, index):
        pass

    def display_rgba_changed(self, display, display_values):
        # when the display rgba data changes, update the display.
        self.update_display_values(display, display_values)

    def display_data_and_metadata_changed(self, display, display_values):
        # when the data changes, no need to do anything. waits for the display rgba to change instead.
        pass

    def update_display_values(self, display, display_values):
        # threadsafe
        data_and_metadata = display.data_and_metadata_for_display_panel
        if data_and_metadata:
            displayed_dimensional_calibrations = display.displayed_dimensional_calibrations
            if len(displayed_dimensional_calibrations) == 0:
                dimensional_calibration = Calibration.Calibration()
            elif len(displayed_dimensional_calibrations) == 1:
                dimensional_calibration = displayed_dimensional_calibrations[0]
            else:
                if data_and_metadata:
                    datum_dimensions = data_and_metadata.datum_dimension_indexes
                    collection_dimensions = data_and_metadata.collection_dimension_indexes
                    if len(collection_dimensions) > 0:
                        dimensional_calibration = data_and_metadata.dimensional_calibrations[collection_dimensions[-1]]
                    elif len(datum_dimensions) > 0:
                        dimensional_calibration = data_and_metadata.dimensional_calibrations[datum_dimensions[-1]]
                    else:
                        dimensional_calibration = Calibration.Calibration()
                else:
                    dimensional_calibration = Calibration.Calibration()

            data_shape = display.preview_2d_shape
            metadata = data_and_metadata.metadata

            # this method may trigger a layout of its parent scroll area. however, the parent scroll
            # area may already be closed. this is a stop-gap guess at a solution - the basic idea being
            # that this object is not closeable while this method is running; and this method should not
            # run if the object is already closed.
            with self.__closing_lock:
                if self.__closed:
                    return

                if self.__image_zoom != display.image_zoom or self.__image_position != display.image_position or self.__image_canvas_mode != display.image_canvas_mode:
                    if display.image_zoom is not None:
                        self.__image_zoom = display.image_zoom
                    if display.image_position is not None:
                        self.__image_position = display.image_position
                    if display.image_canvas_mode is not None:
                        self.__image_canvas_mode = display.image_canvas_mode

                # setting the bitmap on the bitmap_canvas_item is delayed until paint, so that it happens on a thread, since it may be time consuming
                self.__info_overlay_canvas_item.set_data_info(data_shape, dimensional_calibration, metadata)
                # if the data changes, update the display.
                if display_values is not self.__display_values or self.__graphics_changed:
                    self.__graphics_changed = False
                    self.__display_values = display_values
                    self.__data_shape = data_shape
                    if self.__display_frame_rate_id:
                        frame_index = metadata.get("hardware_source", dict()).get("frame_index", 0)
                        if frame_index != self.__display_frame_rate_last_index:
                            Utility.fps_tick("frame_"+self.__display_frame_rate_id)
                            self.__display_frame_rate_last_index = frame_index
                        if id(self.__display_values) != id(self.__last_display_values):
                            Utility.fps_tick("update_"+self.__display_frame_rate_id)
                            self.__last_display_values = self.__display_values
                    # update the cursor info
                    self.__update_cursor_info()

                    def update_layout():
                        # layout. this makes sure that the info overlay gets updated too.
                        with self.update_context():
                            self.__update_image_canvas_size()
                            # trigger updates
                            self.__bitmap_canvas_item.update()
                            with self.__update_layout_handle_lock:
                                self.__update_layout_handle = None

                    if self.__event_loop:
                        with self.__update_layout_handle_lock:
                            update_layout_handle = self.__update_layout_handle
                            if update_layout_handle:
                                update_layout_handle.cancel()
                            scroll_area_canvas_size = self.scroll_area_canvas_item.canvas_size
                            if scroll_area_canvas_size is not None:
                                # only update layout if the size/origin will change. it is slow.
                                image_canvas_origin, image_canvas_size = calculate_origin_and_size(scroll_area_canvas_size, self.__data_shape, self.__image_canvas_mode, self.__image_zoom, self.__image_position)
                                if image_canvas_origin != self.__composite_canvas_item.canvas_origin or image_canvas_size != self.__composite_canvas_item.canvas_size:
                                    self.__update_layout_handle = self.__event_loop.call_soon_threadsafe(update_layout)
                                else:
                                    # trigger updates
                                    self.__bitmap_canvas_item.update()
                                    with self.__update_layout_handle_lock:
                                        update_layout_handle = self.__update_layout_handle
                                        if update_layout_handle:
                                            update_layout_handle.cancel()
                                        self.__update_layout_handle = None

    def update_regions(self, display, graphic_selection):
        self.__graphics = copy.copy(display.graphics)
        self.__graphic_selection = copy.copy(graphic_selection)
        self.__graphics_canvas_item.update_graphics(display.preview_2d_shape, self.__graphics, self.__graphic_selection)
        self.__graphics_changed = True

    def handle_auto_display(self, display) -> bool:
        # enter key has been pressed
        display.auto_display_limits()
        return True

    def __update_image_canvas_zoom(self, new_image_zoom):
        if self.__data_shape is not None:
            self.__image_canvas_mode = "custom"
            self.__image_zoom = new_image_zoom
            self.__update_image_canvas_size()

    # update the image canvas position by the widget delta amount
    def __update_image_canvas_position(self, widget_delta):
        if self.__data_shape is not None:
            # create a widget mapping to get from image norm to widget coordinates and back
            widget_mapping = ImageCanvasItemMapping(self.__data_shape, (0, 0), self.__composite_canvas_item.canvas_size)
            # figure out what composite canvas point lies at the center of the scroll area.
            last_widget_center = widget_mapping.map_point_image_norm_to_widget(self.__image_position)
            # determine what new point will lie at the center of the scroll area by adding delta
            new_widget_center = (last_widget_center[0] + widget_delta[0], last_widget_center[1] + widget_delta[1])
            # map back to image norm coordinates
            new_image_norm_center = widget_mapping.map_point_widget_to_image_norm(new_widget_center)
            # ensure that at least half of the image is always visible
            new_image_norm_center_0 = max(min(new_image_norm_center[0], 1.0), 0.0)
            new_image_norm_center_1 = max(min(new_image_norm_center[1], 1.0), 0.0)
            # save the new image norm center
            self.delegate.update_display_properties({"image_position": list(self.__image_position), "image_canvas_mode": "custom"})
            self.__image_position = (new_image_norm_center_0, new_image_norm_center_1)
            # and update the image canvas accordingly
            self.__image_canvas_mode = "custom"
            self.__update_image_canvas_size()

    # update the image canvas origin and size
    def __update_overlay_canvas_item(self, scroll_area_canvas_size, *, immediate=False):
        image_canvas_origin, image_canvas_size = calculate_origin_and_size(scroll_area_canvas_size, self.__data_shape, self.__image_canvas_mode, self.__image_zoom, self.__image_position)
        if image_canvas_origin is None or image_canvas_size is None:
            self.__info_overlay_canvas_item.image_canvas_origin = None
            self.__info_overlay_canvas_item.image_canvas_size = None
        else:
            self.__composite_canvas_item.update_layout(image_canvas_origin, image_canvas_size, immediate=immediate)
            self.__info_overlay_canvas_item.image_canvas_origin = image_canvas_origin
            self.__info_overlay_canvas_item.image_canvas_size = image_canvas_size

    def __update_image_canvas_size(self):
        scroll_area_canvas_size = self.scroll_area_canvas_item.canvas_size
        if scroll_area_canvas_size is not None:
            self.__update_overlay_canvas_item(scroll_area_canvas_size)
            self.__composite_canvas_item.update()

    def mouse_clicked(self, x, y, modifiers):
        if super().mouse_clicked(x, y, modifiers):
            return True
        # now let the image panel handle mouse clicking if desired
        image_position = self.__get_mouse_mapping().map_point_widget_to_image((y, x))
        return self.delegate.image_clicked(image_position, modifiers)

    def mouse_double_clicked(self, x, y, modifiers):
        self.set_fit_mode()
        return True

    def mouse_pressed(self, x, y, modifiers):
        if super().mouse_pressed(x, y, modifiers):
            return True
        if self.__data_shape is None:
            return False
        image_position = self.__get_mouse_mapping().map_point_widget_to_image((y, x))
        if self.delegate.image_mouse_pressed(image_position, modifiers):
            return True
        self.__undo_command = None
        self.delegate.begin_mouse_tracking()
        # figure out clicked graphic
        self.__graphic_drag_items = []
        self.__graphic_drag_item = None
        self.graphic_drag_item_was_selected = False
        self.__graphic_part_data = {}
        self.__graphic_drag_indexes = []
        if self.delegate.tool_mode == "pointer":
            graphics = self.__graphics
            selection_indexes = self.__graphic_selection.indexes
            start_drag_pos = Geometry.IntPoint(y=y, x=x)
            multiple_items_selected = len(selection_indexes) > 1
            widget_mapping = self.__get_mouse_mapping()
            part_specs = list()
            specific_part_spec = None
            # the graphics are drawn in order, which means the graphics with the higher index are "on top" of the
            # graphics with the lower index. but priority should also be given to selected graphics. so sort the
            # graphics according to whether they are selected or not (selected ones go later), then by their index.
            for graphic_index, graphic in sorted(enumerate(graphics), key=lambda ig: (ig[0] in selection_indexes, ig[0])):
                if isinstance(graphic, (Graphics.PointTypeGraphic, Graphics.LineTypeGraphic, Graphics.RectangleTypeGraphic, Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                    already_selected = graphic_index in selection_indexes
                    move_only = not already_selected or multiple_items_selected
                    try:
                        part, specific = graphic.test(widget_mapping, self.__get_font_metrics_fn, start_drag_pos, move_only)
                    except Exception as e:
                        import traceback
                        logging.debug("Graphic Test Error: %s", e)
                        traceback.print_exc()
                        traceback.print_stack()
                        continue
                    if part:
                        part_spec = graphic_index, graphic, already_selected, "all" if move_only and not part.startswith("inverted") else part
                        part_specs.append(part_spec)
                        if specific:
                            specific_part_spec = part_spec
            # import logging
            # logging.debug(specific_part_spec)
            # logging.debug(part_specs)
            part_spec = specific_part_spec if specific_part_spec is not None else part_specs[-1] if len(part_specs) > 0 else None
            if part_spec is not None:
                graphic_index, graphic, already_selected, part = part_spec
                part = part if specific_part_spec is not None else "all"
                # select item and prepare for drag
                self.graphic_drag_item_was_selected = already_selected
                if not self.graphic_drag_item_was_selected:
                    if modifiers.control:
                        self.delegate.add_index_to_selection(graphic_index)
                        selection_indexes.add(graphic_index)
                    elif not already_selected:
                        self.delegate.set_selection(graphic_index)
                        selection_indexes.clear()
                        selection_indexes.add(graphic_index)
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphics[graphic_index]
                self.__graphic_drag_part = part
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                for index in self.__graphic_drag_indexes:
                    graphic = graphics[index]
                    self.__graphic_drag_items.append(graphic)
                    self.__graphic_part_data[index] = graphic.begin_drag()
            if not self.__graphic_drag_items and not modifiers.control:
                self.delegate.clear_selection()
        elif self.delegate.tool_mode == "line":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_line(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "end"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "rectangle":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_rectangle(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "bottom-right"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "ellipse":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_ellipse(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "bottom-right"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "point":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_point(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "all"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "line-profile":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_line_profile(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "end"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "spot":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            graphic = self.delegate.create_spot(pos)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "bottom-right"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "wedge":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            mouse_angle = math.pi - math.atan2(0.5 - pos[0], 0.5 - pos[1])
            graphic = self.delegate.create_wedge(mouse_angle)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "start-angle"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "ring":
            widget_mapping = self.__get_mouse_mapping()
            pos = widget_mapping.map_point_widget_to_image_norm(Geometry.FloatPoint(y, x))
            radius = math.sqrt((pos[0] - 0.5) ** 2 + (pos[1] - 0.5) ** 2)
            graphic = self.delegate.create_ring(radius)
            self.delegate.add_index_to_selection(self.__graphics.index(graphic))
            if graphic:
                # setup drag
                start_drag_pos = Geometry.IntPoint(y=y, x=x)
                selection_indexes = self.__graphic_selection.indexes
                assert len(selection_indexes) == 1
                self.graphic_drag_item_was_selected = True
                # keep track of general drag information
                self.__graphic_drag_start_pos = start_drag_pos
                self.__graphic_drag_changed = False
                # keep track of info for the specific item that was clicked
                self.__graphic_drag_item = graphic
                self.__graphic_drag_part = "radius_1"
                # keep track of drag information for each item in the set
                self.__graphic_drag_indexes = selection_indexes
                self.__graphic_drag_items.append(graphic)
                self.__graphic_part_data[list(selection_indexes)[0]] = graphic.begin_drag()
                self.__undo_command = self.delegate.create_insert_graphics_command([graphic])
        elif self.delegate.tool_mode == "hand":
            self.__start_drag_pos = (y, x)
            self.__last_drag_pos = (y, x)
            self.__is_dragging = True
        return True

    def mouse_released(self, x, y, modifiers):
        if super().mouse_released(x, y, modifiers):
            return True
        image_position = self.__get_mouse_mapping().map_point_widget_to_image((y, x))
        if self.delegate.image_mouse_released(image_position, modifiers):
            return True
        if self.__data_shape is not None:
            graphics = self.__graphics
            for index in self.__graphic_drag_indexes:
                graphic = graphics[index]
                graphic.end_drag(self.__graphic_part_data[index])
            if self.__graphic_drag_items and not self.__graphic_drag_changed:
                graphic_index = graphics.index(self.__graphic_drag_item)
                # user didn't move graphic
                if not modifiers.control:
                    # user clicked on a single graphic
                    self.delegate.set_selection(graphic_index)
                else:
                    # user control clicked. toggle selection
                    # if control is down and item is already selected, toggle selection of item
                    if self.graphic_drag_item_was_selected:
                        self.delegate.remove_index_from_selection(graphic_index)
                    else:
                        self.delegate.add_index_to_selection(graphic_index)
            self.delegate.end_mouse_tracking(self.__undo_command)
        self.__graphic_drag_items = []
        self.__graphic_drag_item = None
        self.__graphic_part_data = {}
        self.__graphic_drag_indexes = []
        self.__start_drag_pos = None
        self.__last_drag_pos = None
        self.__is_dragging = False
        if self.delegate.tool_mode != "hand":
            self.delegate.tool_mode = "pointer"
        return True

    def mouse_entered(self):
        if super().mouse_entered():
            return True
        self.__mouse_in = True
        return True

    def mouse_exited(self):
        if super().mouse_exited():
            return True
        self.__mouse_in = False
        if self.delegate:  # allow display to work without delegate
            # whenever the cursor exits, clear the cursor display
            self.delegate.cursor_changed(None)
        return True

    def mouse_position_changed(self, x, y, modifiers):
        if super().mouse_position_changed(x, y, modifiers):
            return True
        image_position = self.__get_mouse_mapping().map_point_widget_to_image((y, x))
        if self.delegate.image_mouse_position_changed(image_position, modifiers):
            return True
        if self.delegate.tool_mode == "pointer":
            self.cursor_shape = "arrow"
        elif self.delegate.tool_mode == "line":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "rectangle":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "ellipse":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "point":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "line-profile":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "spot":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "wedge":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "ring":
            self.cursor_shape = "cross"
        elif self.delegate.tool_mode == "hand":
            self.cursor_shape = "hand"
        # x,y already have transform applied
        self.__last_mouse = Geometry.IntPoint(x=x, y=y)
        self.__update_cursor_info()
        if self.__graphic_drag_items:
            if not self.__undo_command:
                self.__undo_command = self.delegate.create_change_graphics_command()
            force_drag = modifiers.only_option
            if force_drag and self.__graphic_drag_part == "all":
                if Geometry.distance(self.__last_mouse, self.__graphic_drag_start_pos) <= 2:
                    self.delegate.drag_graphics(self.__graphic_drag_items)
                    return True
            widget_mapping = self.__get_mouse_mapping()
            self.delegate.update_graphics(widget_mapping, self.__graphic_drag_items, self.__graphic_drag_part,
                                          self.__graphic_part_data, self.__graphic_drag_start_pos,
                                          Geometry.FloatPoint(y=y, x=x), modifiers)
            self.__graphic_drag_changed = True
        elif self.__is_dragging:
            if not self.__undo_command:
                self.__undo_command = self.delegate.create_change_display_command()
            delta = (y - self.__last_drag_pos[0], x - self.__last_drag_pos[1])
            self.__update_image_canvas_position((-delta[0], -delta[1]))
            self.__last_drag_pos = (y, x)
        return True

    def wheel_changed(self, x, y, dx, dy, is_horizontal):
        if self.__mouse_in:
            dx = dx if is_horizontal else 0.0
            dy = dy if not is_horizontal else 0.0
            command = self.delegate.create_change_display_command(command_id="image_position", is_mergeable=True)
            self.__update_image_canvas_position((-dy, -dx))
            self.delegate.push_undo_command(command)
            return True
        return False

    def pan_gesture(self, dx, dy):
        self.__update_image_canvas_position((dy, dx))
        return True

    def context_menu_event(self, x, y, gx, gy):
        return self.delegate.show_display_context_menu(gx, gy)

    # ths message comes from the widget
    def key_pressed(self, key):
        if super().key_pressed(key):
            return True
        # only handle keys if we're directly embedded in an image panel
        if key.is_delete:
            self.delegate.delete_key_pressed()
            return True
        if key.is_enter_or_return:
            self.delegate.enter_key_pressed()
            return True
        if self.__data_shape is not None:
            if self.__graphic_selection.has_selection:
                if key.is_arrow:
                    widget_mapping = self.__get_mouse_mapping()
                    amount = 10.0 if key.modifiers.shift else 1.0
                    if key.is_left_arrow:
                        self.delegate.nudge_selected_graphics(widget_mapping, Geometry.FloatPoint(y=0, x=-amount))
                    elif key.is_up_arrow:
                        self.delegate.nudge_selected_graphics(widget_mapping, Geometry.FloatPoint(y=-amount, x=0))
                    elif key.is_right_arrow:
                        self.delegate.nudge_selected_graphics(widget_mapping, Geometry.FloatPoint(y=0, x=amount))
                    elif key.is_down_arrow:
                        self.delegate.nudge_selected_graphics(widget_mapping, Geometry.FloatPoint(y=amount, x=0))
                    return True
            if key.is_arrow:
                amount = 100.0 if key.modifiers.shift else 10.0
                if key.is_left_arrow:
                    self.move_left(amount)
                elif key.is_up_arrow:
                    self.move_up(amount)
                elif key.is_right_arrow:
                    self.move_right(amount)
                elif key.is_down_arrow:
                    self.move_down(amount)
                return True
            if key.text == "-":
                self.zoom_out()
                return True
            if key.text == "+" or key.text == "=":
                self.zoom_in()
                return True
            if key.text == "j":
                self.move_left()
                return True
            if key.text == "k":
                self.move_right()
                return True
            if key.text == "i":
                self.move_up()
                return True
            if key.text == "m":
                self.move_down()
                return True
            if key.text == "1":
                self.set_one_to_one_mode()
                return True
            if key.text == "2":
                self.set_two_to_one_mode()
                return True
            if key.text == "0":
                self.set_fit_mode()
                return True
            if key.text == ")":
                self.set_fill_mode()
                return True
            if key.key == 70 and key.modifiers.shift and key.modifiers.alt:
                if self.__display_frame_rate_id is None:
                    self.__display_frame_rate_id = str(id(self))
                else:
                    self.__display_frame_rate_id = None
                return True
            if key.key == 76 and key.modifiers.shift and key.modifiers.alt:
                self.__display_latency = not self.__display_latency
                return True
        return False

    def __get_mouse_mapping(self):
        return ImageCanvasItemMapping(self.__data_shape, self.__composite_canvas_item.canvas_origin, self.__composite_canvas_item.canvas_size)

    # map from widget coordinates to image coordinates
    def map_widget_to_image(self, p):
        image_size = self.__data_shape
        transformed_image_rect = self.__get_mouse_mapping().canvas_rect
        if transformed_image_rect and image_size:
            if transformed_image_rect[1][0] != 0.0:
                image_y = image_size[0] * (p[0] - transformed_image_rect[0][0])/transformed_image_rect[1][0]
            else:
                image_y = 0
            if transformed_image_rect[1][1] != 0.0:
                image_x = image_size[1] * (p[1] - transformed_image_rect[0][1])/transformed_image_rect[1][1]
            else:
                image_x = 0
            return image_y, image_x  # c-indexing
        return None

    # map from image normalized coordinates to image coordinates
    def map_image_norm_to_image(self, p):
        image_size = self.__data_shape
        if image_size:
            return p[0] * image_size[0], p[1] * image_size[1]
        return None

    def __update_cursor_info(self):
        if not self.delegate:  # allow display to work without delegate
            return
        image_size = self.__data_shape
        if self.__mouse_in and self.__last_mouse:
            pos_2d = None
            if image_size and len(image_size) > 1:
                pos_2d = self.map_widget_to_image(self.__last_mouse)
            self.delegate.cursor_changed(pos_2d)

    def _inserted(self, container):
        # make sure we get 'prepare_render' calls
        self.register_prepare_canvas_item(self)

    def _removed(self, container):
        # turn off 'prepare_render' calls
        self.unregister_prepare_canvas_item(self)

    def prepare_render(self):
        self.prepare_display()

    def _repaint(self, drawing_context):
        super()._repaint(drawing_context)

        if self.__display_frame_rate_id:
            Utility.fps_tick("display_"+self.__display_frame_rate_id)

        if self.__display_frame_rate_id:
            fps = Utility.fps_get("display_"+self.__display_frame_rate_id)
            fps2 = Utility.fps_get("frame_"+self.__display_frame_rate_id)
            fps3 = Utility.fps_get("update_"+self.__display_frame_rate_id)

            rect = self.canvas_bounds

            drawing_context.save()
            try:
                font = "normal 11px serif"
                text_pos = Geometry.IntPoint(y=rect[0][1], x=rect[0][0])
                drawing_context.begin_path()
                drawing_context.move_to(text_pos.x, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y + 60)
                drawing_context.line_to(text_pos.x, text_pos.y + 60)
                drawing_context.close_path()
                drawing_context.fill_style = "rgba(255, 255, 255, 0.6)"
                drawing_context.fill()
                drawing_context.font = font
                drawing_context.text_baseline = "middle"
                drawing_context.text_align = "left"
                drawing_context.fill_style = "#000"
                drawing_context.fill_text("display:" + fps, text_pos.x + 8, text_pos.y + 10)
                drawing_context.fill_text("frame:" + fps2, text_pos.x + 8, text_pos.y + 30)
                drawing_context.fill_text("update:" + fps3, text_pos.x + 8, text_pos.y + 50)
                drawing_context.statistics("display")
            finally:
                drawing_context.restore()

    # this method will be invoked from the paint thread.
    # data is calculated and then sent to the image canvas item.
    def prepare_display(self):
        if self.__data_shape is not None:
            # configure the bitmap canvas item
            display_values = self.__display_values
            display_data = display_values.display_data_and_metadata
            if display_data and display_data.data.dtype == numpy.float32:
                display_range = display_values.display_range
                color_map_data = display_values.color_map_data
                display_values.finalize()
                if color_map_data is not None:
                    color_map_rgba = numpy.empty(color_map_data.shape[:-1] + (4,), numpy.uint8)
                    color_map_rgba[..., 0:3] = color_map_data
                    color_map_rgba[..., 3] = 255
                    color_map_rgba = color_map_rgba.view(numpy.uint32).reshape(color_map_rgba.shape[:-1])
                else:
                    color_map_rgba = None
                self.__bitmap_canvas_item.set_data(display_data.data, display_range, color_map_rgba, trigger_update=False)
            else:
                data_rgba = display_values.display_rgba
                display_values.finalize()
                if False:
                    # the next section does gaussian blur for image decimation in the case where the destination rectangle
                    # is smaller than the source. this results in lower performance, but higher quality display.
                    height_ratio = (self.__bitmap_canvas_item.canvas_size.height / data_rgba.shape[0]) if data_rgba is not None and data_rgba.shape[0] > 0 else 1
                    if height_ratio < 1:
                        sigma = 0.5 * ((1.0 / height_ratio) - 1.0)
                        data_rgba_copy = numpy.empty_like(data_rgba)
                        data_rgba_u8_view = data_rgba.view(numpy.uint8).reshape(data_rgba.shape + (-1, ))
                        data_rgba_u8_copy_view = data_rgba_copy.view(numpy.uint8).reshape(data_rgba_copy.shape + (-1, ))
                        data_rgba_u8_copy_view[..., 0] = scipy.ndimage.gaussian_filter(data_rgba_u8_view[..., 0], sigma=sigma)
                        data_rgba_u8_copy_view[..., 1] = scipy.ndimage.gaussian_filter(data_rgba_u8_view[..., 1], sigma=sigma)
                        data_rgba_u8_copy_view[..., 2] = scipy.ndimage.gaussian_filter(data_rgba_u8_view[..., 2], sigma=sigma)
                        if data_rgba_u8_view.shape[-1] == 4:
                            data_rgba_u8_copy_view[..., 3] = data_rgba_u8_view[..., 3]
                        data_rgba = data_rgba_copy
                self.__bitmap_canvas_item.set_rgba_bitmap_data(data_rgba, trigger_update=False)
            self.__timestamp_canvas_item.timestamp = display_values.display_rgba_timestamp if self.__display_latency else None

    @property
    def image_canvas_mode(self):
        return self.__image_canvas_mode

    def __apply_display_properties_command(self, display_properties: typing.Mapping):
        command = self.delegate.create_change_display_command()
        self.delegate.update_display_properties(display_properties)
        self.delegate.push_undo_command(command)

    def __apply_move_command(self, delta):
        command = self.delegate.create_change_display_command(command_id="image_nudge", is_mergeable=True)
        self.__update_image_canvas_position(delta)
        self.delegate.push_undo_command(command)

    def set_fit_mode(self):
        self.__apply_display_properties_command({"image_zoom": 1.0, "image_position": (0.5, 0.5), "image_canvas_mode": "fit"})

    def set_fill_mode(self):
        self.__apply_display_properties_command({"image_zoom": 1.0, "image_position": (0.5, 0.5), "image_canvas_mode": "fill"})

    def set_one_to_one_mode(self):
        self.__apply_display_properties_command({"image_zoom": 1.0, "image_position": (0.5, 0.5), "image_canvas_mode": "1:1"})

    def set_two_to_one_mode(self):
        self.__apply_display_properties_command({"image_zoom": 0.5, "image_position": (0.5, 0.5), "image_canvas_mode": "2:1"})

    def zoom_in(self):
        self.__apply_display_properties_command({"image_zoom": self.__image_zoom * 1.25, "image_canvas_mode": "custom"})

    def zoom_out(self):
        self.__apply_display_properties_command({"image_zoom": self.__image_zoom / 1.25, "image_canvas_mode": "custom"})

    def move_left(self, amount=10.0):
        self.__apply_move_command((0.0, amount))

    def move_right(self, amount=10.0):
        self.__apply_move_command((0.0, -amount))

    def move_up(self, amount=10.0):
        self.__apply_move_command((amount, 0.0))

    def move_down(self, amount=10.0):
        self.__apply_move_command((-amount, 0.0))



# standard libraries
import contextlib
import copy
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift.model import DataGroup
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.ui import TestUI


class TestDataGroupClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_deep_copy_should_deep_copy_child_data_groups(self):
        data_group = DataGroup.DataGroup()
        data_item1 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
        data_group.append_data_item(data_item1)
        data_group2 = DataGroup.DataGroup()
        data_group.append_data_group(data_group2)
        # attempt to copy
        data_group_copy = copy.deepcopy(data_group)
        # make sure data_groups are not shared
        self.assertNotEqual(data_group.data_groups[0], data_group_copy.data_groups[0])

    def test_deep_copy_should_not_deep_copy_data_items(self):
        data_group = DataGroup.DataGroup()
        data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
        data_group.append_data_item(data_item)
        data_group_copy = copy.deepcopy(data_group)
        self.assertEqual(data_item.uuid, data_group_copy.data_item_uuids[0])

    def test_counted_data_items(self):
        # TODO: split test_counted_data_items into separate tests
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_group = DataGroup.DataGroup()
            document_model.append_data_group(data_group)
            self.assertEqual(len(data_group.counted_data_items), 0)
            self.assertEqual(len(document_model.data_items), 0)
            data_item1 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item1)
            data_group.append_data_item(data_item1)
            # make sure that both top level and data_group see the data item
            self.assertEqual(len(document_model.data_items), 1)
            self.assertEqual(len(data_group.counted_data_items), 1)
            self.assertEqual(list(document_model.data_items), list(data_group.counted_data_items))
            self.assertIn(data_item1, list(data_group.counted_data_items.keys()))
            # add a child data item and make sure top level and data_group see it
            # also check data item.
            data_item1a = document_model.get_resample_new(data_item1)
            data_group.append_data_item(data_item1a)
            self.assertEqual(len(document_model.data_items), 2)
            self.assertEqual(len(data_group.counted_data_items), 2)
            self.assertIn(data_item1, list(data_group.counted_data_items.keys()))
            self.assertIn(data_item1a, list(data_group.counted_data_items.keys()))
            # add a child data item to the child and make sure top level and data_group match.
            # also check data items.
            data_item1a1 = document_model.get_resample_new(data_item1a)
            display_specifier1a1 = DataItem.DisplaySpecifier.from_data_item(data_item1a1)
            data_group.append_data_item(data_item1a1)
            self.assertEqual(len(document_model.data_items), 3)
            self.assertEqual(len(data_group.counted_data_items), 3)
            self.assertIn(data_item1, list(data_group.counted_data_items.keys()))
            self.assertIn(data_item1a, list(data_group.counted_data_items.keys()))
            self.assertIn(data_item1a1, list(data_group.counted_data_items.keys()))
            # now add a data item that already has children
            data_item2 = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item2)
            data_item2a = document_model.get_resample_new(data_item2)
            data_group.append_data_item(data_item2)
            data_group.append_data_item(data_item2a)
            self.assertEqual(len(document_model.data_items), 5)
            self.assertEqual(len(data_group.counted_data_items), 5)
            self.assertIn(data_item2a, document_model.data_items)
            self.assertIn(data_item2a, list(data_group.counted_data_items.keys()))
            # remove data item without children
            document_model.remove_data_item(data_item1a1)
            self.assertEqual(len(document_model.data_items), 4)
            self.assertEqual(len(data_group.counted_data_items), 4)
            # now remove data item with children
            document_model.remove_data_item(data_item2)
            self.assertEqual(len(document_model.data_items), 2)
            self.assertEqual(len(data_group.counted_data_items), 2)

    def test_deleting_data_item_removes_it_from_data_group(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item)
            document_model.append_data_group(data_group)
            self.assertEqual(1, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            document_model.remove_data_item(data_item)
            self.assertEqual(0, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))

    def test_deleting_data_items_from_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item2 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item1)
            data_group.append_data_item(data_item2)
            data_group.append_data_item(data_item3)
            document_model.append_data_group(data_group)
            # remove two of the three
            command = DocumentController.DocumentController.RemoveDataGroupLibraryItemsCommand(document_model, data_group, [data_item1, data_item3])
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(1, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item2, data_group.data_items[0])
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(3, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item1, data_group.data_items[0])
            self.assertEqual(data_item2, data_group.data_items[1])
            self.assertEqual(data_item3, data_group.data_items[2])
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(1, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item2, data_group.data_items[0])

    def test_deleting_data_items_out_of_order_from_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item2 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item1)
            data_group.append_data_item(data_item2)
            data_group.append_data_item(data_item3)
            document_model.append_data_group(data_group)
            # remove two of the three
            command = DocumentController.DocumentController.RemoveDataGroupLibraryItemsCommand(document_model, data_group, [data_item3, data_item1])
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(1, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item2, data_group.data_items[0])
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(3, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item1, data_group.data_items[0])
            self.assertEqual(data_item2, data_group.data_items[1])
            self.assertEqual(data_item3, data_group.data_items[2])
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(1, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item2, data_group.data_items[0])

    def test_insert_data_item_into_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items and put two in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item2 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item1)
            data_group.append_data_item(data_item3)
            document_model.append_data_group(data_group)
            # insert a new one
            command = document_controller.create_insert_data_group_library_item_command(data_group, 1, data_item2)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(3, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item1, data_group.data_items[0])
            self.assertEqual(data_item2, data_group.data_items[1])
            self.assertEqual(data_item3, data_group.data_items[2])
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(2, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item1, data_group.data_items[0])
            self.assertEqual(data_item3, data_group.data_items[1])
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(3, len(data_group.data_items))
            self.assertEqual(0, len(data_group.data_groups))
            self.assertEqual(data_item1, data_group.data_items[0])
            self.assertEqual(data_item2, data_group.data_items[1])
            self.assertEqual(data_item3, data_group.data_items[2])

    def test_insert_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items and put two in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item3)
            data_group1 = DataGroup.DataGroup()
            data_group1.append_data_item(data_item1)
            data_group2 = DataGroup.DataGroup()
            data_group3 = DataGroup.DataGroup()
            data_group2_uuid = data_group2.uuid
            data_group3.append_data_item(data_item3)
            document_model.append_data_group(data_group1)
            document_model.append_data_group(data_group3)
            # insert the middle data group
            command = DocumentController.DocumentController.InsertDataGroupCommand(document_model, document_model, 1, data_group2)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(3, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[2])
            self.assertEqual(data_group2_uuid, document_model.data_groups[1].uuid)
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(2, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[1])
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(3, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[2])
            self.assertEqual(data_group2_uuid, document_model.data_groups[1].uuid)

    def test_remove_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items and put two in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item3)
            data_group1 = DataGroup.DataGroup()
            data_group1.append_data_item(data_item1)
            data_group2 = DataGroup.DataGroup()
            data_group3 = DataGroup.DataGroup()
            data_group2_uuid = data_group2.uuid
            data_group3.append_data_item(data_item3)
            document_model.append_data_group(data_group1)
            document_model.append_data_group(data_group2)
            document_model.append_data_group(data_group3)
            # remove the middle data group
            command = DocumentController.DocumentController.RemoveDataGroupCommand(document_model, document_model, data_group2)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(2, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[1])
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(3, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[2])
            self.assertEqual(data_group2_uuid, document_model.data_groups[1].uuid)
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(2, len(document_model.data_groups))
            self.assertEqual(data_group1, document_model.data_groups[0])
            self.assertEqual(data_group3, document_model.data_groups[1])

    def test_data_group_rename_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup data group
            data_group = DataGroup.DataGroup()
            data_group.title = "ethel"
            document_model.append_data_group(data_group)
            # rename
            command = document_controller.create_rename_data_group_command(data_group, "fred")
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(data_group.title, "fred")
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(data_group.title, "ethel")
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(data_group.title, "fred")

    def test_data_item_removed_implicitly_from_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items and put in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item2 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            document_model.append_data_item(data_item3)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item1)
            data_group.append_data_item(data_item2)
            data_group.append_data_item(data_item3)
            document_model.append_data_group(data_group)
            # remove the 2nd data item
            command = document_controller.create_remove_library_items_command([data_item2])
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(data_group.data_items))
            self.assertEqual(document_model.data_items[0], data_group.data_items[0])
            self.assertEqual(document_model.data_items[1], data_group.data_items[1])
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(3, len(document_model.data_items))
            self.assertEqual(3, len(data_group.data_items))
            self.assertEqual(document_model.data_items[0], data_group.data_items[0])
            self.assertEqual(document_model.data_items[1], data_group.data_items[1])
            self.assertEqual(document_model.data_items[2], data_group.data_items[2])
            # redo and check
            document_controller.handle_redo()
            self.assertEqual(2, len(document_model.data_items))
            self.assertEqual(2, len(data_group.data_items))
            self.assertEqual(document_model.data_items[0], data_group.data_items[0])
            self.assertEqual(document_model.data_items[1], data_group.data_items[1])

    def test_inserting_items_data_group_undo_redo(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # setup three data items and put in a group
            data_item1 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item2 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item3 = DataItem.DataItem(numpy.random.randn(4, 4))
            data_item4 = DataItem.DataItem(numpy.random.randn(4, 4))
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item3)
            document_model.append_data_item(data_item4)
            data_group = DataGroup.DataGroup()
            data_group.append_data_item(data_item1)
            data_group.append_data_item(data_item3)
            document_model.append_data_group(data_group)
            self.assertEqual(3, len(document_model.data_items))
            self.assertEqual(2, len(data_group.data_items))
            self.assertListEqual([data_item1, data_item3, data_item4], list(document_model.data_items))
            self.assertListEqual([data_item1, data_item3], list(data_group.data_items))
            # insert new items
            command = DocumentController.DocumentController.InsertDataGroupLibraryItemsCommand(document_controller, data_group, [data_item2, data_item4], 1)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(4, len(document_model.data_items))
            self.assertEqual(4, len(data_group.data_items))
            self.assertListEqual([data_item1, data_item3, data_item4, data_item2], list(document_model.data_items))
            self.assertListEqual([data_item1, data_item2, data_item4, data_item3], list(data_group.data_items))
            # undo and check
            document_controller.handle_undo()
            self.assertEqual(3, len(document_model.data_items))
            self.assertEqual(2, len(data_group.data_items))
            self.assertListEqual([data_item1, data_item3, data_item4], list(document_model.data_items))
            self.assertListEqual([data_item1, data_item3], list(data_group.data_items))
            # redo and check
            document_controller.handle_redo()
            data_item2 = document_model.data_items[3]
            self.assertEqual(4, len(document_model.data_items))
            self.assertEqual(4, len(data_group.data_items))
            self.assertListEqual([data_item1, data_item3, data_item4, data_item2], list(document_model.data_items))
            self.assertListEqual([data_item1, data_item2, data_item4, data_item3], list(data_group.data_items))


if __name__ == '__main__':
    unittest.main()
